# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Support code to adapt existing `VGG pytorch model <vgg-pytorch_>`_ functionality to U-Net style network for segmentation."""

import torchvision.models
from torch.hub import load_state_dict_from_url


class VGG4Segmentation(torchvision.models.vgg.VGG):
    """Adaptation of `VGG pytorch model <vgg-pytorch_>`_ functionality to U-Net style network for segmentation.

    This version of VGG is slightly modified so it can be used through
    torchvision's API.  It outputs intermediate features which are normally not
    output by the base VGG implementation, but are required for segmentation
    operations.

    Parameters
    ----------
    *args
        Arguments to be passed to the parent VGG model.
    **kwargs
        Keyword arguments to be passed to the parent VGG model.

        * ``return_features`` (:py:class:`list`): An optional list of integers indicating
          the feature layers to be returned from the original module.
    """

    def __init__(self, *args, **kwargs):
        self._return_features = kwargs.pop("return_features")
        super().__init__(*args, **kwargs)

    def forward(self, x):
        outputs = []
        # hardwiring of input
        outputs.append(x.shape[2:4])
        for index, m in enumerate(self.features):
            x = m(x)
            # extract layers
            if index in self._return_features:
                outputs.append(x)
        return outputs


def _make_vgg16_type_d_for_segmentation(pretrained, batch_norm, progress, **kwargs):
    if pretrained:
        kwargs["init_weights"] = False

    model = VGG4Segmentation(
        torchvision.models.vgg.make_layers(
            torchvision.models.vgg.cfgs["D"],
            batch_norm=batch_norm,
        ),
        **kwargs,
    )

    if pretrained:
        weights = (
            torchvision.models.vgg.VGG16_Weights.DEFAULT.url
            if not batch_norm
            else torchvision.models.vgg.VGG16_BN_Weights.DEFAULT.url
        )

        state_dict = load_state_dict_from_url(weights, progress=progress)
        model.load_state_dict(state_dict)

    # erase VGG head (for classification), not used for segmentation
    delattr(model, "classifier")
    delattr(model, "avgpool")

    return model


def vgg16_for_segmentation(pretrained=False, progress=True, **kwargs):
    """Create an instance of VGG16.

    Parameters
    ----------
    pretrained
        If True, usees VGG16 pretrained weights.
    progress
        If True, shows a progress bar when downloading weights.
    **kwargs
        Keyword arguments to be passed to the parent VGG model.

        * ``return_features`` (:py:class:`list`): An optional list of integers indicating
          the feature layers to be returned from the original module.

    Returns
    -------
        Instance of VGG16.
    """
    return _make_vgg16_type_d_for_segmentation(
        pretrained=pretrained, batch_norm=False, progress=progress, **kwargs
    )


vgg16_for_segmentation.__doc__ = torchvision.models.vgg16.__doc__


def vgg16_bn_for_segmentation(pretrained=False, progress=True, **kwargs):
    """Create an instance of VGG16 with batch norm.

    Parameters
    ----------
    pretrained
        If True, usees VGG16 pretrained weights.
    progress
        If True, shows a progress bar when downloading weights.
    **kwargs
        Keyword arguments to be passed to the parent VGG model.

        * ``return_features`` (:py:class:`list`): An optional list of integers indicating
          the feature layers to be returned from the original module.

    Returns
    -------
        Instance of VGG16.
    """
    return _make_vgg16_type_d_for_segmentation(
        pretrained=pretrained, batch_norm=True, progress=progress, **kwargs
    )


vgg16_bn_for_segmentation.__doc__ = torchvision.models.vgg16_bn.__doc__



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Definition of types related to models used in classification tasks."""

import typing

Prediction: typing.TypeAlias = tuple[str, typing.Sequence[int], typing.Sequence[float]]
"""The sample name, the target, and the predicted value."""

PredictionSplit: typing.TypeAlias = typing.Mapping[str, typing.Sequence[Prediction]]
"""A series of predictions for different database splits."""

SaliencyMapAlgorithm: typing.TypeAlias = typing.Literal[
    "ablationcam",
    "eigencam",
    "eigengradcam",
    "fullgrad",
    "gradcam",
    "gradcamelementwise",
    "gradcam++",
    "gradcamplusplus",
    "hirescam",
    "layercam",
    "randomcam",
    "scorecam",
    "xgradcam",
]
"""Supported saliency map algorithms."""



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Faster R-CNN object detection (and classification) network architecture, from :cite:p:`ren_faster_2017`."""

import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.detect.faster_rcnn
import mednet.models.losses
import mednet.models.transforms

model = mednet.models.detect.faster_rcnn.FasterRCNN(
    optimizer_type=torch.optim.SGD,
    optimizer_arguments=dict(lr=0.005, momentum=0.9, weight_decay=0.0005),
    scheduler_type=torch.optim.lr_scheduler.StepLR,
    scheduler_arguments=dict(step_size=3, gamma=0.1),
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
    pretrained=True,
    num_classes=1,
    variant="mobilenetv3-small",  # fastest testing
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click
from clapper.click import ResourceOption, verbosity_option

from ...click import ConfigCommand
from ...logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Evaluate the generated saliency maps for their localization performance:

   .. code:: sh

      mednet classify saliency interpretability -vv pasa tbx11k-v1-healthy-vs-atb --input-folder=parent-folder/saliencies/ --output-json=path/to/interpretability-scores.json

""",
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance implementing the network architecture
    (not the weights, necessarily) to be used for inference.  Currently, only
    supports pasa and densenet models.""",
    required=True,
    type=click.UNPROCESSED,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="""A lightning DataModule that will be asked for prediction data
    loaders. Typically, this includes all configured splits in a DataModule,
    however this is not a requirement.  A DataModule that returns a single
    dataloader for prediction (wrapped in a dictionary) is acceptable.""",
    required=True,
    type=click.UNPROCESSED,
    cls=ResourceOption,
)
@click.option(
    "--input-folder",
    "-i",
    help="""Path from where to load saliency maps.  You can generate saliency
    maps with ``mednet classify saliency generate``.""",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
    ),
    default="saliency-maps",
    cls=ResourceOption,
)
@click.option(
    "--target-label",
    "-t",
    help="""The target label that will be analysed.  It must match the target
    label that was used to generate the saliency maps provided with option
    ``--input-folder``.  Samples with all other labels are ignored.""",
    required=True,
    type=click.INT,
    default=1,
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="""Path to the folder in which all measures will be saved.""",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
        writable=True,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--only-dataset",
    "-S",
    help="""If set, will only run the command for the named dataset on the
    provided datamodule, skipping any other dataset.""",
    cls=ResourceOption,
)
@click.option(
    "--plot/--no-plot",
    "-P",
    help="""If set, then also produces figures containing the plots of
    score histograms.""",
    required=True,
    show_default=True,
    default=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def interpretability(
    model,
    datamodule,
    input_folder,
    target_label,
    output_folder,
    only_dataset,
    plot: bool,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Evaluate saliency map agreement with annotations (human
    interpretability).

    The evaluation happens by comparing saliency maps with ground-truth
    provided by any other means (typically following a manual annotation
    procedure).

    .. note::

       For obvious reasons, this evaluation is limited to datasets that
       contain built-in annotations which corroborate classification.

    As a result of the evaluation, this application creates a single .json file
    that resembles the original DataModule, with added information containing
    the following measures, for each sample:

    * Proportional Energy: A measure that compares (UNthresholed) saliency maps
      with annotations (based on :cite:p:`wang_score-cam_2020`). It estimates how much
      activation lies within the ground truth boxes compared to the total sum
      of the activations.

    * Average Saliency Focus: estimates how much of the ground truth bounding
      boxes area is covered by the activations.  It is similar to the
      proportional energy measure in the sense that it does not need explicit
      thresholding.
    """

    import matplotlib.backends.backend_pdf

    from ....engine.classify.saliency.interpretability import run
    from ....engine.classify.saliency.utils import make_plots, make_table
    from ....scripts.utils import save_json_metadata, save_json_with_backup

    datamodule.model_transforms = list(model.model_transforms)
    datamodule.batch_size = 1
    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    output_json = output_folder / "interpretability.json"

    # stores all information we can think of, to reproduce this later
    save_json_metadata(
        output_file=output_json.with_suffix(".meta.json"),
        model=model,
        datamodule=datamodule,
        output_json=output_json,
        input_folder=input_folder,
        target_label=target_label,
        only_dataset=only_dataset,
        plot=plot,
    )

    results = run(
        input_folder=input_folder,
        target_label=target_label,
        datamodule=datamodule,
        only_dataset=only_dataset,
    )

    logger.info(f"Saving output file to `{str(output_json)}`...")
    save_json_with_backup(output_json, results)

    table = make_table(
        results=results, indexes={2: "PropEng", 3: "AvgSalFocus"}, format_="rst"
    )
    output_table = output_json.with_suffix(".rst")
    logger.info(f"Saving output summary table to `{str(output_table)}`...")
    with output_table.open("w") as f:
        f.write(table)
    click.echo(table)

    # Plots histograms, if the user asked to do so.
    if plot:
        figure_path = output_json.with_suffix(".pdf")
        logger.info(f"Saving plots to `{str(figure_path)}`...")
        with matplotlib.backends.backend_pdf.PdfPages(figure_path) as pdf:
            for fig in make_plots(
                results=results,
                indexes={2: "Proportional Energy", 3: "Average Saliency Focus"},
            ):
                pdf.savefig(fig)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for PadChest dataset."""

import importlib

import pytest
from click.testing import CliRunner


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "split,lengths",
    [
        # ("idiap.json.bz2", dict(train=96269)),  ## many labels
        ("tb-idiap.json", dict(train=200, test=50)),  # 0: no-tb, 1: tb
        (
            "no-tb-idiap.json.bz2",
            dict(train=54371, validation=4052),
        ),  # 14 labels
        ("cardiomegaly-idiap.json", dict(train=40)),  # 14 labels
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lengths: dict[str, int],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.classify.data.padchest", split),
        lengths=lengths,
        prefixes=("",),
        possible_labels=(0, 1),
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.padchest")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["--limit=10", "padchest-idiap"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


testdata = [
    ("idiap", "train", 193),
    ("idiap", "test", 1),
    ("tb_idiap", "train", 1),
    ("no_tb_idiap", "train", 14),
    ("cardiomegaly_idiap", "train", 14),
]


@pytest.mark.skip_if_rc_var_not_set("datadir.padchest")
@pytest.mark.parametrize("name,dataset,num_labels", testdata)
def test_loading(database_checkers, name: str, dataset: str, num_labels: int):
    datamodule = importlib.import_module(
        f".{name}",
        "mednet.config.classify.data.padchest",
    ).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.setup("predict")  # sets up all datasets

    if dataset in datamodule.predict_dataloader():
        loader = datamodule.predict_dataloader()[dataset]

        limit = 3  # limit load checking
        for batch in loader:
            if limit == 0:
                break
            database_checkers.check_loaded_batch(
                batch,
                batch_size=1,
                color_planes=1,
                prefixes=("",),
                possible_labels=(0, 1),
                expected_num_labels=num_labels,
                expected_meta_size=2,
            )
            limit -= 1


@pytest.mark.skip_if_rc_var_not_set("datadir.padchest")
def test_loaded_image_quality(database_checkers, datadir):
    reference_histogram_file = (
        datadir / "histograms/raw_data/histograms_padchest_idiap.json"
    )

    datamodule = importlib.import_module(
        ".idiap", "mednet.config.classify.data.padchest"
    ).datamodule

    datamodule.model_transforms = []
    datamodule.setup("predict")

    database_checkers.check_image_quality(datamodule, reference_histogram_file)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Definition of base model type for object detection tasks."""

import logging
import typing

import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data

from ...data.typing import TransformSequence
from ..model import Model as BaseModel

logger = logging.getLogger(__name__)


class Model(BaseModel):
    """Base model type for semantic segmentation tasks.

    Parameters
    ----------
    name
        Common name to give to models of this type.
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        name: str,
        loss_type: type[torch.nn.Module] | None = None,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        num_classes: int = 1,
    ):
        super().__init__(
            name=name,
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""U-Net for image segmentation.

U-Net is a convolutional neural network that was developed for biomedical image
segmentation at the Computer Science Department of the University of Freiburg,
Germany.  The network is based on the fully convolutional network (FCN) and its
architecture was modified and extended to work with fewer training images and
to yield more precise segmentations.

Reference: :cite:p:`navab_u-net_2015`
"""

import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.losses
import mednet.models.segment.unet
import mednet.models.transforms

model = mednet.models.segment.unet.Unet(
    loss_type=mednet.models.losses.SoftJaccardAndBCEWithLogitsLoss,
    loss_arguments=dict(alpha=0.7),  # 0.7 BCE + 0.3 Jaccard
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=0.01),
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
    pretrained=False,
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Lightning callbacks to log custom measurements."""

import logging
import time
import typing
import warnings

import lightning.pytorch
import lightning.pytorch.callbacks
import torch

from ..utils.resources import ResourceMonitor, aggregate

logger = logging.getLogger(__name__)


class LoggingCallback(lightning.pytorch.Callback):
    """Callback to log various training metrics and device information.

    Rationale:

    1. Losses are logged at the end of every batch, accumulated and handled by
       the lightning framework.
    2. Everything else is done at the end of a training or validation epoch and
       mostly concerns runtime metrics such as memory and cpu/gpu utilisation.

    Parameters
    ----------
    resource_monitor
        A monitor that watches resource usage (CPU/GPU) in a separate process
        and totally asynchronously with the code execution.
    """

    def __init__(
        self,
        resource_monitor: ResourceMonitor,
    ):
        super().__init__()

        # timers
        self._start_training_time = time.time()
        self._start_training_epoch_time = time.time()
        self._start_validation_epoch_time = time.time()

        # log accumulators for a single flush at each training cycle
        self._to_log: dict[str, float] = {}

        # helpers for CPU and GPU utilisation
        self._resource_monitor = resource_monitor
        self._max_queue_retries = 2

    def on_train_start(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ):
        """Execute actions when training starts (lightning callback).

        This method is executed whenever you *start* training a module.

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """
        self._start_training_time = time.time()

    def on_train_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ):
        """Execute actions when training ends (lightning callback).

        This method is executed whenever you *end* training a module.

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """
        pass

    def on_train_epoch_start(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ) -> None:
        """Execute actions when a training epoch starts (lightning callback).

        This method is executed whenever a training epoch starts.  Presumably,
        batches happen as often as possible.  You want to make this code very
        fast.  Do not log things to the terminal or the such, or do complicated
        (lengthy) calculations.

        .. warning::

           This is executed **while** you are training.  Be very succint or
           face the consequences of slow training!

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """

        # summarizes resource usage since the last checkpoint
        # clears internal buffers and starts accumulating again.
        self._resource_monitor.clear()

        self._start_training_epoch_time = time.time()

    def on_train_epoch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ):
        """Execute actions after a training epoch ends (lightning callback).

        This method is executed whenever a training epoch ends.  Presumably,
        epochs happen as often as possible.  You want to make this code
        relatively fast to avoid significative runtime slow-downs.

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """

        # evaluates this training epoch total time, and log it
        epoch_time = time.time() - self._start_training_epoch_time

        self._to_log["epoch-duration-seconds/train"] = epoch_time

        if len(pl_module.trainer.lr_scheduler_configs) == 0:
            self._to_log["learning-rate"] = pl_module.optimizers().defaults["lr"]  # type: ignore
        else:
            self._to_log["learning-rate"] = pl_module.trainer.lr_scheduler_configs[
                0
            ].scheduler.get_last_lr()[0]  # type: ignore

        overall_cycle_time = time.time() - self._start_training_epoch_time
        self._to_log["cycle-time-seconds/train"] = overall_cycle_time
        self._to_log["total-execution-time-seconds"] = (
            time.time() - self._start_training_time
        )
        self._to_log["eta-seconds"] = overall_cycle_time * (
            trainer.max_epochs - trainer.current_epoch  # type: ignore
        )
        # the "step" is the tensorboard jargon for "epoch" or "batch",
        # depending on how we are logging - in a more general way, it simply
        # means the relative time step.
        self._to_log["step"] = float(trainer.current_epoch)

        # Do not log during sanity check as results are not relevant
        if not trainer.sanity_checking:
            pl_module.log_dict(self._to_log)
            self._to_log = {}

    def on_train_batch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
        outputs: typing.Mapping[str, torch.Tensor],
        batch: typing.Mapping[str, typing.Any],
        batch_idx: int,
    ) -> None:
        """Execute actions after a training batch ends (lightning callback).

        This method is executed whenever a training batch ends.  Presumably,
        batches happen as often as possible.  You want to make this code very
        fast.  Do not log things to the terminal or the such, or do complicated
        (lengthy) calculations.

        .. warning::

           This is executed **while** you are training.  Be very succint or
           face the consequences of slow training!

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        outputs
            The outputs of the module's ``training_step``.
        batch
            The data that the training step received.
        batch_idx
            The relative number of the batch.
        """

        batch_size = batch["image"].shape[0]

        pl_module.log(
            "loss/train",
            outputs["loss"].item(),
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            batch_size=batch_size,
        )

    def on_validation_epoch_start(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ) -> None:
        """Execute actions before a validation batch starts (lightning callback).

        This method is executed whenever a validation batch starts.  Presumably,
        batches happen as often as possible.  You want to make this code very
        fast.  Do not log things to the terminal or the such, or do complicated
        (lengthy) calculations.

        .. warning::

           This is executed **while** you are training.  Be very succint or
           face the consequences of slow training!

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """

        current_timestamp = time.time()

        # required because the validation epoch is started **within** the
        # training epoch START/END.
        # summarizes resource usage since the last checkpoint
        metrics = self._resource_monitor.checkpoint()

        # filter out samples not acquired during the training epoch
        metrics = [
            k
            for k in metrics
            if k["timestamp"] > self._start_training_epoch_time
            and k["timestamp"] < current_timestamp
        ]

        if not metrics:
            warnings.warn(
                f"CPU/GPU monitor worker did not provide any measures. "
                f"This can be due to the length of the monitoring interval, "
                f"currently set to {self._resource_monitor.interval}s, "
                f"compared to the time I've been accumulating measurements "
                f"({(time.time()-self._start_training_epoch_time):.1f}s). "
                f"A way this can be fixed is by reducing the monitoring "
                f"interval to a suitable value, so it allows some measures "
                f"to be performed. Note this is only possible if the time "
                f"to log a single measurement point is smaller than the "
                f"time it takes to **train** a single epoch."
            )
        else:
            for metric_name, metric_value in aggregate(metrics).items():
                self._to_log[f"{metric_name}/train"] = float(metric_value)

        self._start_validation_epoch_time = time.time()
        self._resource_monitor.clear()

    def on_validation_epoch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
    ) -> None:
        """Execute actions after a validation batch ends (lightning callback).

        This method is executed whenever a validation epoch ends.  Presumably,
        epochs happen as often as possible.  You want to make this code
        relatively fast to avoid significative runtime slow-downs.

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        """

        current_timestamp = time.time()

        # summarizes resource usage since the last checkpoint
        # clears internal buffers and starts accumulating again.
        metrics = self._resource_monitor.checkpoint()

        # filter out samples not acquired during the validation epoch
        metrics = [
            k
            for k in metrics
            if k["timestamp"] > self._start_training_epoch_time
            and k["timestamp"] < current_timestamp
        ]

        if not metrics:
            warnings.warn(
                f"CPU/GPU monitor worker did not provide any measures. "
                f"This can be due to the length of the monitoring interval, "
                f"currently set to {self._resource_monitor.interval}s, "
                f"compared to the time I've been accumulating measurements "
                f"({(time.time()-self._start_validation_epoch_time):.1f}s). "
                f"A way this can be fixed is by reducing the monitoring "
                f"interval to a suitable value, so it allows some measures "
                f"to be performed. Note this is only possible if the time "
                f"to log a single measurement point is smaller than the "
                f"time it takes to **validate** a single epoch."
            )
        else:
            for metric_name, metric_value in aggregate(metrics).items():
                self._to_log[f"{metric_name}/validation"] = float(metric_value)

        epoch_time = time.time() - self._start_validation_epoch_time
        self._to_log["epoch-duration-seconds/validation"] = epoch_time

        self._to_log["step"] = float(trainer.current_epoch)

        # Do not log during sanity check as results are not relevant
        if not trainer.sanity_checking:
            pl_module.log_dict(self._to_log)
            self._to_log = {}

    def on_validation_batch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
        outputs: torch.Tensor,
        batch: typing.Mapping[str, typing.Any],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Execute actions after a validation after ends (lightning callback).

        This method is executed whenever a validation batch ends.  Presumably,
        batches happen as often as possible.  You want to make this code very
        fast.  Do not log things to the terminal or the such, or do complicated
        (lengthy) calculations.

        .. warning::

           This is executed **while** you are training.  Be very succint or
           face the consequences of slow training!

        Parameters
        ----------
        trainer
            The Lightning trainer object.
        pl_module
            The lightning module that is being trained.
        outputs
            The outputs of the module's ``training_step``.
        batch
            The data that the training step received.
        batch_idx
            The relative number of the batch.
        dataloader_idx
            Index of the dataloader used during validation.  Use this to figure
            out which dataset was used for this validation epoch.
        """

        if dataloader_idx == 0:
            key = "loss/validation"
        else:
            key = f"loss/validation-{dataloader_idx}"

        batch_size = batch["image"].shape[0]

        pl_module.log(
            key,
            outputs.item(),
            prog_bar=False,
            on_step=False,
            on_epoch=True,
            batch_size=batch_size,
            add_dataloader_idx=False,
        )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Montgomery database transformed for lung detection.

Check :py:mod:`.segment.montgomery` for details.  This module only uses the
segmentation utilities to provide an "object detection" interface.

* Output sample:

    * Image: As per :py:mod:`.segment.montgomery`.
    * Bounding-box: A single bounding-box accounting for the observed lung region.

This module contains the base declaration of common data modules and raw-data loaders
for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import pathlib
import typing

import torch
from torchvision import tv_tensors
from torchvision.ops import masks_to_boxes

from ..datamodule import CachingDataModule
from ..segment.montgomery import CONFIGURATION_KEY_DATADIR, DATABASE_SLUG
from ..segment.montgomery import RawDataLoader as BaseDataLoader
from ..split import JSONDatabaseSplit
from ..typing import Sample


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the montgomery dataset."""

    def __init__(self):
        super().__init__()

        # keep this on this module for correct database script support!
        assert CONFIGURATION_KEY_DATADIR

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        retval = super().sample(sample)

        target = tv_tensors.BoundingBoxes(
            data=self.target(sample),
            format=tv_tensors.BoundingBoxFormat.XYXY,
            canvas_size=retval["image"].shape[-2:],
        )

        return dict(
            image=retval["image"],
            target=target,
            labels=torch.FloatTensor([1]),  # background is 0
            mask=retval["mask"],
            name=sample[0],
        )

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        # converts target into bounding box
        return masks_to_boxes(super().target(sample))


class DataModule(CachingDataModule):
    """Montgomery database transformed for lung detection.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="detection",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import copy
import logging
import typing

import lightning.pytorch
import torch
import torch.nn
import torch.optim.lr_scheduler
import torch.optim.optimizer
import torch.utils.data
import torchvision.transforms

from ..data.datamodule import ConcatDataModule
from ..data.typing import TransformSequence
from ..utils.string import rewrap
from .typing import Checkpoint

logger = logging.getLogger(__name__)


class Model(lightning.pytorch.LightningModule):
    """Base class for models.

    Parameters
    ----------
    name
        Common name to give to models of this type.
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed to the
           batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        name: str,
        loss_type: type[torch.nn.Module] | None = None,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        num_classes: int = 1,
    ):
        super().__init__()

        self.name = name
        self.loss_type = loss_type
        self.train_loss_arguments = copy.deepcopy(loss_arguments or {})
        self.validation_loss_arguments = copy.deepcopy(loss_arguments or {})
        self._optimizer_type = optimizer_type
        self._optimizer_arguments = copy.deepcopy(optimizer_arguments or {})
        self._scheduler_type = scheduler_type
        self._scheduler_arguments = copy.deepcopy(scheduler_arguments or {})
        self.model_transforms = copy.deepcopy(model_transforms or [])
        self.augmentation_transforms = copy.deepcopy(augmentation_transforms or [])
        self._num_classes = num_classes

        # controls normalization and changes to the normalizer
        self._default_normalizer: typing.Callable[[torch.Tensor], torch.Tensor] = (
            lambda x: x
        )
        self._normalizer: typing.Callable[[torch.Tensor], torch.Tensor] = (
            self._default_normalizer
        )

        # initializes losses from input arguments
        self.configure_losses()

    @property
    def augmentation_transforms(self) -> torchvision.transforms.Compose:
        return self._augmentation_transforms

    @augmentation_transforms.setter
    def augmentation_transforms(self, v: TransformSequence) -> None:
        self._augmentation_transforms = torchvision.transforms.Compose(v)

        if len(v) != 0:
            transforms_str = ", ".join(
                [
                    f"{type(k).__module__}.{str(k)}"
                    for k in self._augmentation_transforms.transforms
                ]
            )
            logger.info(f"Data augmentations: {transforms_str}")
        else:
            logger.info("Data augmentations: None")

    @property
    def num_classes(self) -> int:
        """Number of outputs (classes) for this model.

        Returns
        -------
        int
            The number of outputs supported by this model.
        """
        return self._num_classes

    @num_classes.setter
    def num_classes(self, v: int) -> None:
        # on this top-level class, we do not know how to affect the number of output
        # classes supported by all possible models.
        if self._num_classes != v:
            raise NotImplementedError(
                f"Cannot reset number of classes to `{v}` for model `{self.name}`."
            )

    @property
    def normalizer(self) -> typing.Callable[[torch.Tensor], torch.Tensor]:
        """Normalizer for input images.

        Returns
        -------
            Callable (typically a :py:class:`torch.nn.Module`) that takes the input
            tensor to be normalized and returns its normalized version.
        """
        return self._normalizer

    @normalizer.setter
    def normalizer(self, v: typing.Callable[[torch.Tensor], torch.Tensor]) -> None:
        if self._normalizer != self._default_normalizer:
            raise RuntimeError(
                "Attempt to reset input normalizer after non-default initialization."
            )
        self._normalizer = v

    def normalizer_is_set(self) -> bool:
        """Tell if a normalizer different than the default (NOOP) was set.

        Returns
        -------
            ``True`` if a normalizer different than the default (NOOP) was set.  Else,
            returns ``False``.
        """
        return self._normalizer != self._default_normalizer

    def set_normalizer_from_dataloader(
        self, dataloader: torch.utils.data.DataLoader
    ) -> None:
        """Initialize the input normalizer for the current model.

        Sets-up a z-normalization scheme based on the input dataloader samples.

        Parameters
        ----------
        dataloader
            A torch Dataloader from which to compute the mean and std.
        """
        from .normalizer import make_z_normalizer

        logger.info("Computing z-norm input normalization from dataloader.")
        self._normalizer = make_z_normalizer(dataloader)

    def on_save_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Perform actions during checkpoint saving (called by lightning).

        Called by Lightning when saving a checkpoint to give you a chance to store
        anything else you might want to save. Use on_load_checkpoint() to restore what
        additional data is saved here.

        Parameters
        ----------
        checkpoint
            The checkpoint to save.
        """

        if self.normalizer_is_set():
            # we need to save as this is NOT the NOOP function we typically attach
            checkpoint["normalizer"] = self._normalizer

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Perform actions during model loading (called by lightning).

        If you saved something with on_save_checkpoint() this is your chance to restore
        this.

        Parameters
        ----------
        checkpoint
            The loaded checkpoint.
        """

        normalizer = checkpoint.get("normalizer")
        if normalizer is None:
            self._normalizer = self._default_normalizer
        else:
            self._normalizer = normalizer
            logger.info("Restored input normalizer from checkpoint.")

        super().on_load_checkpoint(typing.cast(typing.Any, checkpoint))

    def balance_losses(self, datamodule: ConcatDataModule) -> None:
        """Balance the loss based on the distribution of positives.

        This function will balance the loss with considering the targets in the
        datamodule. Only works if the loss supports it (i.e. contains a ``pos_weight``
        attribute).

        Parameters
        ----------
        datamodule
            Instance of a datamodule from where targets will be loaded.
        """

        if self.loss_type is None:
            logger.info("No loss configured to be balanced.")
            return

        if issubclass(self.loss_type, torch.nn.BCEWithLogitsLoss):
            logger.info("Applying train/valid loss balancing...")
            from .losses import pos_weight_for_bcewithlogitsloss

            # special case for BCEWithLogitsLoss, which requires configuration
            train_loss_args, valid_loss_args = pos_weight_for_bcewithlogitsloss(
                datamodule
            )
            self.train_loss_arguments.update(train_loss_args)
            self.validation_loss_arguments.update(valid_loss_args)

        elif hasattr(self.loss_type, "get_arguments_from_datamodule"):
            logger.info("Applying train/valid loss balancing...")
            # it is one of our custom losses that knows how to self-configure
            train_loss_args, valid_loss_args = (
                self.loss_type.get_arguments_from_datamodule(datamodule)
            )
            self.train_loss_arguments.update(train_loss_args)
            self.validation_loss_arguments.update(valid_loss_args)

        else:
            logger.warning(
                rewrap(
                    f"""Loss `{self.loss_type}` is not supported and will not
                    be balanced."""
                )
            )

        # calls super class to continue the loss configuration
        self.configure_losses()

    def configure_losses(self) -> None:
        """Create loss objects for train and validation."""

        if self.loss_type is None:
            logger.debug("No loss to configure.")
            return

        logger.info(f"Configuring train loss ({self.train_loss_arguments})...")
        self.train_loss = self.loss_type(**self.train_loss_arguments)
        logger.info(
            f"Configuring validation loss ({self.validation_loss_arguments})..."
        )
        self.validation_loss = self.loss_type(**self.validation_loss_arguments)

    def configure_optimizers(self):
        """Configure optimizers.

        Returns
        -------
            The configured optimizer, or a tuple containing the optimizer and a
            scheduler, in case one is present.
        """

        optimizer = self._optimizer_type(self.parameters(), **self._optimizer_arguments)

        if self._scheduler_type is None:
            logger.debug("No scheduler to be configured.")
            return optimizer

        scheduler = self._scheduler_type(optimizer, **self._scheduler_arguments)
        return [optimizer], [scheduler]

    def to(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Self:
        """Move model, augmentations and losses to specified device.

        Refer to the method :py:meth:`torch.nn.Module.to` for details.

        Parameters
        ----------
        *args
            Parameter forwarded to the underlying implementations.
        **kwargs
            Parameter forwarded to the underlying implementations.

        Returns
        -------
            Self.
        """

        super().to(*args, **kwargs)

        self._augmentation_transforms = torchvision.transforms.Compose(
            [
                k.to(*args, **kwargs)
                for k in self._augmentation_transforms.transforms
                if hasattr(k, "to")
            ]
        )

        if self.loss_type is not None:
            self.train_loss.to(*args, **kwargs)
            self.validation_loss.to(*args, **kwargs)

        return self

    def forward(self, *args, **kwargs):
        del args, kwargs  # satisfies linter
        raise NotImplementedError

    def training_step(self, batch, batch_idx):
        del batch, batch_idx  # satisfies linter
        raise NotImplementedError(
            "You cannot use the base model without implementing the training_step()"
        )

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        del batch_idx, dataloader_idx  # satisfies linter

        # debug code to inspect images by eye:
        # from torchvision.transforms.v2.functional import to_pil_image
        # for k in batch["image"]:
        #    to_pil_image(k).show()
        #    __import__("pdb").set_trace()

        return self.validation_loss(self(batch["image"]), batch["target"])

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        del batch_idx, dataloader_idx  # satisfies linter

        # debug code to inspect images by eye:
        # from torchvision.transforms.v2.functional import to_pil_image
        # for k in batch["image"]:
        #    to_pil_image(k).show()
        #    __import__("pdb").set_trace()

        return torch.sigmoid(self(batch["image"]))



# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of :py:mod:`montgomery's <.data.classify.montgomery>` and :py:mod:`shenzhen's <.data.classify.shenzhen>` splits.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import pathlib

from ..datamodule import ConcatDataModule
from ..split import JSONDatabaseSplit
from .montgomery import RawDataLoader as MontgomeryLoader
from .shenzhen import RawDataLoader as ShenzhenLoader

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of :py:mod:`montgomery's <.data.classify.montgomery>` and :py:mod:`shenzhen's <.data.classify.shenzhen>` splits.

    Parameters
    ----------
    split_name
        The name of the split to assign to this data module.
    split_path
        Path or traversable (resource) with the JSON split description to load
        for montgomery and shenzhen databases (in this order).
    """

    def __init__(
        self,
        split_name: str,
        split_path: tuple[
            pathlib.Path | importlib.resources.abc.Traversable,
            pathlib.Path | importlib.resources.abc.Traversable,
        ],
    ):
        montgomery_loader = MontgomeryLoader()
        montgomery_split = JSONDatabaseSplit(split_path[0])
        shenzhen_loader = ShenzhenLoader()
        shenzhen_split = JSONDatabaseSplit(split_path[1])

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                ],
            },
            database_name=DATABASE_SLUG,
            split_name=split_name,
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""TBX11k database for TB detection.

* Database reference: :cite:p:`liu_rethinking_2020`
* The original database contains samples of healthy, sick (no TB), active
  and latent TB cases.  There is a total of 11702 samples in the database.
  Healthy and sick individuals are kept in separate folders.  Latent and
  active TB cases are merged in the same directory.  One must check the
  radiological annotations to understand if samples contain either, or both
  (latent and active TB) signs.
* There is one case of patient (file ``imgs/tb/tb1199.png``), that is
  inside the ``tb`` folder, but contains no annotations.  This sample was
  **excluded** from our splits.
* There are 30 cases of patients that have both active and latent TB
  radiological signs, over the entire database.  Those samples were also
  **excluded** from our splits:

  - imgs/tb/tb0135.png
  - imgs/tb/tb0142.png
  - imgs/tb/tb0154.png
  - imgs/tb/tb0167.png
  - imgs/tb/tb0190.png
  - imgs/tb/tb0246.png
  - imgs/tb/tb0255.png
  - imgs/tb/tb0279.png
  - imgs/tb/tb0284.png
  - imgs/tb/tb0350.png
  - imgs/tb/tb0378.png
  - imgs/tb/tb0392.png
  - imgs/tb/tb0395.png
  - imgs/tb/tb0501.png
  - imgs/tb/tb0506.png
  - imgs/tb/tb0526.png
  - imgs/tb/tb0543.png
  - imgs/tb/tb0639.png
  - imgs/tb/tb0640.png
  - imgs/tb/tb0667.png
  - imgs/tb/tb0676.png
  - imgs/tb/tb0713.png
  - imgs/tb/tb0786.png
  - imgs/tb/tb0870.png
  - imgs/tb/tb0875.png
  - imgs/tb/tb0945.png
  - imgs/tb/tb0949.png
  - imgs/tb/tb0968.png
  - imgs/tb/tb1104.png
  - imgs/tb/tb1143.png

* Original train database samples:

  - Healthy: 3000
  - Sick (but no TB): 3000
  - Active TB only: 473
  - Latent TB only: 103
  - Both active and latent TB: 23
  - Unknown: 1
  - Total: 6600

* Original validation database samples:

  - Healthy: 800
  - Sick (but no TB): 800
  - Latent TB only: 36
  - Active TB only: 157
  - Both active and latent TB: 7
  - Total: 1800

* Original test database samples:

  - Unknown: 3302
  - Total: 3302

* Because the test set does not have annotations, we generated train,
  validation and test databases as such:

  - The original validation database becomes our test set.
  - The original train database is split into new train and validation
    splits (validation ratio = 0.203 w.r.t. original train database size).
    The selection of samples is stratified (see comments through our split
    code, which is shipped alongside this file.)

.. important:: **Raw data organization**

    The TBX11k_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least these two
    subdirectories:

    - ``imgs/`` (directory containing sub-directories and images in PNG format)
    - ``annotations/`` (directory containing labels in JSON and XML format)

Data specifications:

* Raw data input (on disk): PNG images 8 bits RGB, 512 x 512 pixels

* Output image:

  * Transforms:

    - Load raw PNG with :py:mod:`PIL`
    - Convert to torch tensor

  * Final specifications:

    - RGB, encoded as a 3-plane tensor using 32-bit floats, square
      (512x512 pixels)
    - Labels: 0 (healthy, latent tb or sick but no tb depending on the
      protocol), 1 (active tuberculosis), as a torch float tensor.
    - Bounding-boxes: indicating regions of the image that corroborate (active
      or latent TB diagnostics).

.. note:: **JSON Encoding**

   Details of the encoding of database splits in JSON format.

   For healthy/sick (no TB)/latent TB cases, each sample is represented by a
   filename, relative to the root of the installed database, followed by the
   number 0 (negative class).

   For active TB cases, each sample is represented by a filename, followed by
   the number 1, and then by 1 or more 5-tuples with radiological finding
   locations, as described above.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import copy
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
import torch.utils.data
import torchvision.transforms.v2
import torchvision.tv_tensors

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


DatabaseSample: typing.TypeAlias = (
    tuple[str, int] | tuple[str, int, tuple[tuple[int, int, int, int, int]]]
)
"""Type of objects in our JSON representation for this database.

For healthy/sick (no TB)/latent TB cases, each sample is represented by
a filename, relative to the root of the installed database, followed by
the number 0 (negative class).

For active TB cases, each sample is represented by a filename, followed
by the number 1, and then by 1 or more 5-tuples with radiological
finding locations, as described above.
"""


def custom_collate_fn(batch):  # numpydoc ignore=PR01
    """Collate samples that include bounding boxes.

    This allows us to have :py:class:`torchvision.tv_tensors.BoundingBoxes`
    that can contain zero to multiple boxes, which is not supported by the
    default collate function that uses :py:func:`torch.stack` for batching.

    Returns
    -------
        The given batch.
    """

    # A copy of the batch is needed, otherwise this function will permamnently
    # modify data in case it is cached.
    batch_ = copy.deepcopy(batch)

    if "bboxes" not in batch_[0].keys():
        return torch.utils.data.default_collate(batch_)

    # Remove the BoundingBoxes from the batch and apply the default collate
    # function
    bboxes = []
    bboxes_targets = []

    [bboxes.append(b.pop("bboxes")) for b in batch_]
    [bboxes_targets.append(b.pop("bbox_targets")) for b in batch_]

    out = torch.utils.data.default_collate(batch_)

    # Insert the BoundingBoxes and targets back as lists
    out["bboxes"] = bboxes
    out["bbox_targets"] = bboxes_targets

    return out


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the TBX11k database.

    Parameters
    ----------
    ignore_bboxes
        If True, sample() does not return bounding boxes.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self, ignore_bboxes: bool = False):
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )
        self.ignore_bboxes = ignore_bboxes

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the database root folder,
            where to find the image to be loaded, an integer, representing the
            sample target, and possible radiological findings represented by
            bounding boxes.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0])
        image = torchvision.transforms.v2.functional.to_dtype(
            torchvision.transforms.v2.functional.to_image(image),
            torch.float32,
            scale=True,
        )
        image = torchvision.tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        if self.ignore_bboxes:
            return dict(image=image, target=self.target(sample), name=sample[0])

        bboxes, bbox_targets = self.bounding_boxes(
            sample, (image.shape[-2], image.shape[-1])
        )

        return dict(
            image=image,
            target=self.target(sample),
            name=sample[0],
            bboxes=bboxes,
            bbox_targets=bbox_targets,
        )

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            1D torch float tensor.
        """

        return torch.FloatTensor([sample[1]])

    def bounding_boxes(
        self, sample: typing.Any, canvas_size: tuple[int, int]
    ) -> tuple[torch.Tensor | None, torch.Tensor | None]:
        """Load image annotated bounding-boxes from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the database root folder,
            where to find the image to be loaded, an integer, representing the
            sample target, and possible radiological findings represented by
            bounding boxes.
        canvas_size
            Size of the full image.

        Returns
        -------
            Bounding box annotations, if any available with the sample, and
            their targets (1 for active TB, 0 for latent TB region).
        """

        if len(sample) > 2:
            bbox_data = [s[1:] for s in sample[2]]
            bboxes = torchvision.transforms.v2.functional.convert_bounding_box_format(
                torchvision.tv_tensors.BoundingBoxes(
                    data=bbox_data,
                    format=torchvision.tv_tensors.BoundingBoxFormat.XYWH,
                    canvas_size=canvas_size,
                ),
                new_format=torchvision.tv_tensors.BoundingBoxFormat.XYXY,
            )
            bbox_targets = [s[0] for s in sample[2]]
            return bboxes, torch.Tensor(bbox_targets).byte()

        return None, None


class DataModule(CachingDataModule):
    """TBX11k database for TB detection.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    ignore_bboxes
        If True, sample() does not return bounding boxes.
    """

    def __init__(
        self,
        split_path: pathlib.Path | importlib.resources.abc.Traversable,
        ignore_bboxes: bool = False,
    ):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(ignore_bboxes=ignore_bboxes),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            collate_fn=custom_collate_fn,
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Pre-configured database splits and models for classification tasks."""



# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""String manipulation utilities."""


def rewrap(s: str) -> str:
    """Re-wrap input string to remove all whitespaces.

    This function will transform all whitespaces (one or more spaces, tabs, newlines,
    returns, formfeeds) into a single space on a given string.

    Parameters
    ----------
    s
        The input string.

    Returns
    -------
        The "re-wrapped" string with multiple whitespaces transformed to a single space.
    """
    return " ".join(s.split())



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Engine and functions for score completeness analysis."""

import functools
import logging
import multiprocessing
import typing

import lightning.pytorch
import numpy as np
import torch
import tqdm
from pytorch_grad_cam.metrics.road import (
    ROADLeastRelevantFirstAverage,
    ROADMostRelevantFirstAverage,
)
from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget

from ....data.typing import Sample
from ....engine.device import DeviceManager
from ....models.classify.typing import SaliencyMapAlgorithm

logger = logging.getLogger(__name__)


class SigmoidClassifierOutputTarget(torch.nn.Module):
    """Consider output to be a sigmoid.

    Parameters
    ----------
    category
        The category.
    """

    def __init__(self, category):
        self.category = category

    def __call__(self, model_output):
        sigmoid_output = torch.sigmoid(model_output)
        if len(sigmoid_output.shape) == 1:
            return sigmoid_output[self.category]
        return sigmoid_output[:, self.category]


def _calculate_road_scores(
    model: lightning.pytorch.LightningModule,
    images: torch.Tensor,
    output_num: int,
    saliency_map_callable: typing.Callable,
    percentiles: typing.Sequence[int],
) -> tuple[float, float, float]:
    """Calculate average ROAD scores for different removal percentiles.

    This function calculates ROAD scores by averaging the scores for
    different removal (hardcoded) percentiles, for a single input image, a
    given visualization method, and a target class.

    Parameters
    ----------
    model
        Neural network model (e.g. pasa).
    images
        A batch of input images to use for evaluating the ROAD scores.  Currently,
        we only support batches with a single image.
    output_num
        Target output neuron to take into consideration when evaluating the
        saliency maps and calculating ROAD scores.
    saliency_map_callable
        A callable saliency-map generator from grad-cam.
    percentiles
        A sequence of percentiles (percent x100) integer values indicating the
        proportion of pixels to perturb in the original image to calculate both
        MoRF and LeRF scores.

    Returns
    -------
    tuple[float, float, float]
        A 3-tuple containing floating point numbers representing the
        most-relevant-first average score (``morf``), least-relevant-first
        average score (``lerf``) and the combined value (``(lerf-morf)/2``).
    """

    saliency_map = saliency_map_callable(
        input_tensor=images,
        targets=[ClassifierOutputTarget(output_num)],
    )

    cam_metric_roadmorf_avg = ROADMostRelevantFirstAverage(
        percentiles=percentiles,
    )
    cam_metric_roadlerf_avg = ROADLeastRelevantFirstAverage(
        percentiles=percentiles,
    )

    # Calculate ROAD scores for all percentiles and average - this is NOT the
    # current processing bottleneck.  If you want to optimise anyting, look at
    # the evaluation of the perturbation using scipy.sparse at the
    # NoisyLinearImputer, part of the grad-cam package (submodule
    # ``metrics.road``).
    metric_target = [SigmoidClassifierOutputTarget(output_num)]

    morf_scores = cam_metric_roadmorf_avg(
        input_tensor=images,
        cams=saliency_map,
        model=model,
        targets=metric_target,
    )

    lerf_scores = cam_metric_roadlerf_avg(
        input_tensor=images,
        cams=saliency_map,
        model=model,
        targets=metric_target,
    )

    return (
        float(morf_scores.item()),
        float(lerf_scores.item()),
        float(lerf_scores.item() - morf_scores.item()) / 2.0,
    )


def _process_sample(
    sample: Sample,
    model: lightning.pytorch.LightningModule,
    device: torch.device,
    saliency_map_callable: typing.Callable,
    target_class: typing.Literal["highest", "all"],
    positive_only: bool,
    percentiles: typing.Sequence[int],
) -> list:
    """Process a single sample.

    Helper function to :py:func:`run` to be used in multiprocessing contexts.

    Parameters
    ----------
    sample
        The Sample to process.
    model
        Neural network model (e.g. pasa).
    device
        The device to process samples on.
    saliency_map_callable
        A callable saliency-map generator from grad-cam.
    target_class
        Class to target for saliency estimation. Can be set to
        "all" or "highest". "highest" is default, which means
        only saliency maps for the class with the highest
        activation will be generated.
    positive_only
        If set, and the model chosen has a single output (binary), then
        saliency maps will only be generated for samples of the positive class.
    percentiles
        A sequence of percentiles (percent x100) integer values indicating the
        proportion of pixels to perturb in the original image to calculate both
        MoRF and LeRF scores.

    Returns
    -------
    list
        A list containing the following items for a particular sample:
        * The relative path to the sample.
        * The label.
        * An index to the specified target_class.
        * The computed ROAD scores.
    """

    name: str = sample["name"][0]
    label: int = int(sample["target"].item())
    image = sample["image"]

    # in binary classification systems, negative labels may be skipped
    if positive_only and (model.num_classes == 1) and (label == 0):
        return [name, label]

    # chooses target outputs to generate saliency maps for
    if model.num_classes > 1:  # type: ignore
        if target_class == "all":
            # test all outputs
            for output_num in range(model.num_classes):  # type: ignore
                results = _calculate_road_scores(
                    model,
                    image,
                    output_num,
                    saliency_map_callable,
                    percentiles,
                )
                return [name, label, output_num, *results]

        else:
            # we will figure out the output with the highest value and
            # evaluate the saliency mapping technique over it.
            outputs = saliency_map_callable.activations_and_grads(image)  # type: ignore
            output_nums = np.argmax(outputs.cpu().data.numpy(), axis=-1)
            assert len(output_nums) == 1
            results = _calculate_road_scores(
                model,
                image,
                output_nums[0],
                saliency_map_callable,
                percentiles,
            )
            return [name, label, output_nums[0], *results]

    # default route for binary classification
    results = _calculate_road_scores(
        model,
        image,
        0,
        saliency_map_callable,
        percentiles,
    )
    return [name, label, 0, *results]


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
    saliency_map_algorithm: SaliencyMapAlgorithm,
    target_class: typing.Literal["highest", "all"],
    positive_only: bool,
    percentiles: typing.Sequence[int],
    parallel: int,
    only_dataset: str | None,
) -> dict[str, list[typing.Any]]:
    """Evaluate ROAD scores for all samples in a DataModule.

    The ROAD algorithm was first described in :cite:p:`rong_consistent_2022`. It estimates
    explainability (in the completeness sense) of saliency maps by substituting
    relevant pixels in the input image by a local average, re-running
    prediction on the altered image, and measuring changes in the output
    classification score when said perturbations are in place.  By substituting
    the most or least relevant pixels with surrounding averages, the ROAD algorithm
    estimates the importance of such elements in the produced saliency map.  As
    of 2023, this measurement technique is considered to be one of the
    state-of-the-art metrics of explainability.

    This function returns a dictionary containing most-relevant-first (remove a
    percentile of the most relevant pixels), least-relevant-first (remove a
    percentile of the least relevant pixels), and combined ROAD evaluations per
    sample for a particular saliency mapping algorithm.

    Parameters
    ----------
    model
        Neural network model (e.g. pasa).
    datamodule
        The lightning DataModule to iterate on.
    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a lightning accelerator setup.
    saliency_map_algorithm
        The algorithm for saliency map estimation to use.
    target_class
        (Use only with multi-label models) Which class to target for CAM
        calculation. Can be set to "all" or "highest". "highest" is
        default, which means only saliency maps for the class with the highest
        activation will be generated.
    positive_only
        If set, saliency maps will only be generated for positive samples (ie.
        label == 1 in a binary classification task).  This option is ignored on
        a multi-class output model.
    percentiles
        A sequence of percentiles (percent x100) integer values indicating the
        proportion of pixels to perturb in the original image to calculate both
        MoRF and LeRF scores.
    parallel
        Use multiprocessing for data processing: if set to -1, disables
        multiprocessing.  Set to 0 to enable as many data processing instances
        as processing cores available in the system.  Set to >= 1 to enable
        that many multiprocessing instances for data processing.
    only_dataset
        If set, will only run this code for the named dataset on the provided
        datamodule, skipping any other datasets.

    Returns
    -------
    dict[str, list[typing.Any]]
        A dictionary where keys are dataset names in the provide DataModule,
        and values are lists containing sample information alongside metrics
        calculated:

        * Sample name
        * Sample target class
        * The model output number used for the ROAD analysis (0, for binary
          classifers as there is typically only one output).
        * ``morf``: ROAD most-relevant-first average of percentiles 20, 40, 60 and
          80 (a.k.a. AOPC-MoRF).
        * ``lerf``: ROAD least-relevant-first average of percentiles 20, 40, 60 and
          80 (a.k.a. AOPC-LeRF).
        * combined: Average ROAD combined score by evaluating ``(lerf-morf)/2``
          (a.k.a. AOPC-Combined).
    """

    from ....models.classify.densenet import Densenet
    from ....models.classify.pasa import Pasa
    from .generator import _create_saliency_map_callable

    if isinstance(model, Pasa):
        if saliency_map_algorithm == "fullgrad":
            raise ValueError(
                "Fullgrad saliency map algorithm is not supported for the "
                "Pasa model.",
            )
        target_layers = [model.fc14]  # Last non-1x1 Conv2d layer
    elif isinstance(model, Densenet):
        target_layers = [
            model.model.features.denseblock4.denselayer16.conv2,  # type: ignore
        ]
    else:
        raise TypeError(f"Model of type `{type(model)}` is not yet supported.")

    if device_manager.device_type in ("cuda", "mps") and (
        parallel == 0 or parallel > 1
    ):
        raise RuntimeError(
            f"The number of multiprocessing instances is set to {parallel} and "
            f"you asked to use a GPU (device = `{device_manager.device_type}`"
            f"). The current implementation can only handle a single GPU.  "
            f"Either disable GPU usage, set the number of "
            f"multiprocessing instances to one, or disable multiprocessing "
            "entirely (ie. set it to -1).",
        )

    # prepares model for evaluation, cast to target device
    device = device_manager.torch_device()
    model.eval()

    saliency_map_callable = _create_saliency_map_callable(
        saliency_map_algorithm,
        model,
        target_layers,  # type: ignore
    )

    retval: dict[str, list[typing.Any]] = {}

    # our worker function
    _process = functools.partial(
        _process_sample,
        model=model,
        device=device,
        saliency_map_callable=saliency_map_callable,
        target_class=target_class,
        positive_only=positive_only,
        percentiles=percentiles,
    )

    for k, v in datamodule.predict_dataloader().items():
        if only_dataset is not None and k != only_dataset:
            logger.warning(
                f"Skipping processing for dataset `{k}` following user request..."
            )
            continue

        retval[k] = []

        if parallel < 0:
            logger.info(
                f"Computing ROAD scores for dataset `{k}` in the current "
                f"process context...",
            )
            for sample in tqdm.tqdm(
                v,
                desc="samples",
                leave=False,
                disable=None,
            ):
                retval[k].append(_process(sample))

        else:
            instances = parallel or multiprocessing.cpu_count()
            logger.info(
                f"Computing ROAD scores for dataset `{k}` using {instances} "
                f"processes...",
            )
            with multiprocessing.Pool(instances) as p:
                retval[k] = list(tqdm.tqdm(p.imap(_process, v), total=len(v)))

    return retval



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for refuge dataset."""

import importlib

import pytest
from click.testing import CliRunner


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "split,lengths",
    [
        ("default", dict(train=400, validation=400, test=400)),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lengths: dict[str, int],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.segment.data.refuge", f"{split}.json"),
        lengths=lengths,
        prefixes=["Training400/", "REFUGE-Validation400/V", "Test400/T0"],
        possible_labels=[],
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.refuge")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["--limit=10", "refuge-disc"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


@pytest.mark.skip_if_rc_var_not_set("datadir.refuge")
@pytest.mark.parametrize(
    "dataset",
    ["train", "validation", "test"],
)
@pytest.mark.parametrize(
    "name",
    ["cup", "disc"],
)
def test_loading(database_checkers, name: str, dataset: str):
    datamodule = importlib.import_module(
        f".{name}",
        "mednet.config.segment.data.refuge",
    ).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.setup("predict")  # sets up all datasets

    loader = datamodule.predict_dataloader()[dataset]

    limit = 3  # limit load checking
    for batch in loader:
        if limit == 0:
            break
        database_checkers.check_loaded_batch(
            batch,
            batch_size=1,
            color_planes=3,
            expected_num_labels=1,
            expected_meta_size=3,
            prefixes=["Training400/", "REFUGE-Validation400/V", "Test400/T0"],
            possible_labels=[],
        )
        limit -= 1


@pytest.mark.skip_if_rc_var_not_set("datadir.refuge")
def test_raw_transforms_image_quality(database_checkers, datadir):
    reference_histogram_file = (
        datadir / "histograms/raw_data/histograms_refuge_cup.json"
    )

    datamodule = importlib.import_module(
        ".cup", "mednet.config.segment.data.refuge"
    ).datamodule

    datamodule.model_transforms = []
    datamodule.setup("predict")

    database_checkers.check_image_quality(datamodule, reference_histogram_file)


@pytest.mark.skip_if_rc_var_not_set("datadir.refuge")
@pytest.mark.parametrize(
    "model_name",
    ["lwnet"],
)
def test_model_transforms_image_quality(database_checkers, datadir, model_name):
    reference_histogram_file = (
        datadir / f"histograms/models/histograms_{model_name}_refuge_cup.json"
    )

    datamodule = importlib.import_module(
        ".cup", "mednet.config.segment.data.refuge"
    ).datamodule

    model = importlib.import_module(
        f".{model_name}", "mednet.config.segment.models"
    ).model

    datamodule.model_transforms = model.model_transforms
    datamodule.setup("predict")

    database_checkers.check_image_quality(
        datamodule,
        reference_histogram_file,
        compare_type="statistical",
        pearson_coeff_threshold=0.005,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib

import click
from clapper.click import AliasedGroup


@click.group(
    cls=AliasedGroup,
    context_settings=dict(help_option_names=["-?", "-h", "--help"]),
)
def cli():
    """Medical image analysis AI toolbox."""
    pass


def _add_command(module, obj):
    cli.add_command(
        getattr(importlib.import_module("." + module, package=__name__), obj)
    )


_add_command(".info", "info")
_add_command(".info", "version")
_add_command(".config", "config")
_add_command(".database", "database")
_add_command(".train", "train")
_add_command(".train_analysis", "train_analysis")
_add_command(".predict", "predict")
_add_command(".experiment", "experiment")
_add_command(".upload", "upload")
_add_command(".classify.cli", "classify")
_add_command(".detect.cli", "detect")
_add_command(".segment.cli", "segment")



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Functions to compute normalisation factors based on dataloaders."""

import logging

import torch
import torch.nn
import torch.utils.data
import torchvision.transforms
import tqdm

logger = logging.getLogger(__name__)


def make_z_normalizer(
    dataloader: torch.utils.data.DataLoader,
) -> torchvision.transforms.Normalize:
    """Compute mean and standard deviation from a dataloader.

    This function will input a dataloader, and compute the mean and standard
    deviation by image channel.  It will work for both monochromatic, and color
    inputs with 2, 3 or more color planes.

    Parameters
    ----------
    dataloader
        A torch Dataloader from which to compute the mean and std.

    Returns
    -------
        An initialized normalizer.
    """

    # Peek the number of channels of batches in the data loader
    batch = next(iter(dataloader))
    channels = batch["image"].shape[1]

    # Initialises accumulators
    mean = torch.zeros(channels, dtype=batch["image"].dtype)
    var = torch.zeros(channels, dtype=batch["image"].dtype)
    num_images = 0

    # Evaluates mean and standard deviation
    for batch in tqdm.tqdm(dataloader, unit="batch"):
        data = batch["image"]
        data = data.view(data.size(0), data.size(1), -1)

        num_images += data.size(0)
        mean += data.mean(2).sum(0)
        var += data.var(2).sum(0)

    mean /= num_images
    var /= num_images
    std = torch.sqrt(var)

    return torchvision.transforms.Normalize(mean, std)


def make_imagenet_normalizer() -> torchvision.transforms.Normalize:
    """Return the stock ImageNet normalisation weights from torchvision.

    The weights are wrapped in a torch module.  This normalizer only works for
    **RGB (color) images**.

    Returns
    -------
        An initialized normalizer.
    """

    return torchvision.transforms.Normalize(
        (0.485, 0.456, 0.406),
        (0.229, 0.224, 0.225),
    )


def make_cocov1_normalizer() -> torchvision.transforms.Normalize:
    """Return the stock COCO v1 normalisation weights from torchvision.

    The weights are wrapped in a torch module.  This normalizer only works for
    **RGB (color) images**.

    Returns
    -------
        An initialized normalizer.
    """

    return torchvision.transforms.Normalize(
        (0.0, 0.0, 0.0),
        (1.0, 1.0, 1.0),
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tools for interacting with the running computer or GPU."""

import logging
import multiprocessing
import multiprocessing.synchronize
import os
import plistlib
import queue
import shutil
import subprocess
import time
import typing
import warnings

import numpy
import psutil

from ..engine.device import SupportedPytorchDevice

logger = logging.getLogger(__name__)

_nvidia_smi = shutil.which("nvidia-smi")
"""Location of the nvidia-smi program, if one exists."""


GB = float(2**30)
"""The number of bytes in a gigabyte."""


def _run_nvidia_smi(query: typing.Sequence[str]) -> dict[str, str | float]:
    """Return GPU information from query.

    For a comprehensive list of options and help, execute ``nvidia-smi
    --help-query-gpu`` on a host with a GPU

    Parameters
    ----------
    query
        A list of query strings as defined by ``nvidia-smi --help-query-gpu``.

    Returns
    -------
    dict[str, str | float]
        A dictionary containing the queried parameters (``rename`` versions).
        If ``nvidia-smi`` is not available, returns ``None``.  Percentage
        information is left alone, memory information is transformed to
        gigabytes (floating-point).
    """

    if _nvidia_smi is None:
        warnings.warn(
            "Cannot find `nvidia-smi` on your $PATH. Install it or choose "
            "another backend to run. Cannot return GPU information without "
            "this executable."
        )
        return {}

    # Gets GPU information, based on a GPU device if that is set. Returns
    # ordered results.
    query_str = f"{_nvidia_smi} --query-gpu={','.join(query)} --format=csv,noheader"
    visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES")
    if visible_devices:
        query_str += f" --id={visible_devices}"
    values = subprocess.getoutput(query_str)

    retval: dict[str, str | float] = {}
    for i, k in enumerate([k.strip() for k in values.split(",")]):
        retval[query[i]] = k
        if k.endswith("%"):
            retval[query[i]] = float(k[:-1].strip())
        elif k.endswith("MiB"):
            retval[query[i]] = float(k[:-3].strip()) / 1024
    return retval


def _run_powermetrics(
    time_window_ms: int = 500, key: str | None = None
) -> dict[str, typing.Any]:
    """Return GPU information from the system.

    For a comprehensive list of options and help, execute ``man powermetrics``
    on a Mac computer with Apple silicon.

    Parameters
    ----------
    time_window_ms
        The amount of time, in milliseconds, to collect usage information on
        the GPU.
    key
        If specified returns only a sub-key of the dictionary.

    Returns
    -------
        A dictionary containing the GPU information.
    """

    cmd = [
        "sudo",
        "-n",
        "/usr/bin/powermetrics",
        "--samplers",
        "gpu_power",
        f"-i{time_window_ms}",
        "-n1",
        "-fplist",
    ]

    try:
        raw_bytes = subprocess.check_output(cmd)
        data = plistlib.loads(raw_bytes)
        return data[key] if key is not None else data

    except subprocess.CalledProcessError:
        warnings.warn(
            f"Cannot run `sudo powermetrics` without a password. Probably, "
            f"you did not setup sudo to execute the macOS CLI application "
            f"`powermetrics` passwordlessly and therefore this warning is "
            f"being issued. This does not affect the running of your model "
            f"training, only the ability of the resource monitor of gathering "
            f"GPU usage information on macOS while using the MPS compute "
            f"backend.  To configure this properly and get rid of this "
            f"warning, execute `sudo visudo` and add the following line where "
            f"suitable: `yourusername ALL=(ALL) NOPASSWD:SETENV: "
            f"/usr/bin/powermetrics`. Replace `yourusername` by your actual "
            f"username on the machine. Test the setup running the command "
            f"`{' '.join(cmd)}` by hand."
        )

    return {}


def cuda_constants() -> dict[str, str | int | float]:
    """Return GPU (static) information using nvidia-smi.

    Check ``_run_nvidia_smi()`` for operational details if necessary.

    Returns
    -------
    dict[str, str | int | float]
        If ``nvidia-smi`` is not available, returns ``None``, otherwise, we
        return a dictionary containing the following ``nvidia-smi`` query
        information, in this order:

        * ``gpu_name``, as ``gpu_name`` (:py:class:`str`)
        * ``driver_version``, as ``gpu_driver_version`` (:py:class:`str`)
        * ``memory.total``, as ``gpu_memory_total`` (transformed to gigabytes,
          :py:class:`float`)
    """

    retval = _run_nvidia_smi(("gpu_name", "driver_version", "memory.total"))
    if retval:
        retval["driver-version/gpu"] = retval.pop("driver_version")
        retval["total-memory-GB/gpu"] = retval.pop("memory.total")
    return retval


def mps_constants() -> dict[str, str | int | float]:
    """Return GPU (static) information using `/usr/bin/powermetrics`.

    Returns
    -------
    dict[str, str | int | float]
        If ``nvidia-smi`` is not available, returns ``None``, otherwise, we
        return a dictionary containing the following ``nvidia-smi`` query
        information, in this order:

        * ``gpu_name``, as ``gpu_name`` (:py:class:`str`)
        * ``driver_version``, as ``gpu_driver_version`` (:py:class:`str`)
        * ``memory.total``, as ``gpu_memory_total`` (transformed to gigabytes,
          :py:class:`float`)
    """

    raw_bytes = subprocess.check_output(
        ["/usr/sbin/system_profiler", "-xml", "SPDisplaysDataType"],
    )
    data = plistlib.loads(raw_bytes)
    name = data[0]["_items"][0]["_name"]
    no_gpu_cores = int(data[0]["_items"][0]["sppci_cores"])

    return {
        "apple-processor-model": name,
        "number-of-cores/gpu": no_gpu_cores,
    }


def _cuda_log() -> dict[str, float]:
    """Return GPU information about current non-static status using nvidia-
    smi.

    See :py:func:`_run_nvidia_smi` for operational details.

    Returns
    -------
    dict[str, float]
        If ``nvidia-smi`` is not available, returns ``None``, otherwise, we
        return a dictionary containing the following ``nvidia-smi`` query
        information, in this order:

        * ``memory.used``, as ``memory-used-GB/gpu`` (transformed to gigabytes,
          :py:class:`float`)
        * ``memory.free``, as ``memory-free-GB/gpu`` (transformed to gigabytes,
          :py:class:`float`)
        * ``100*memory.used/memory.total``, as ``memory-percent/gpu``,
          (:py:class:`float`, in percent)
        * ``utilization.gpu``, as ``percent-usage/gpu``,
          (:py:class:`float`, in percent)
    """

    result = _run_nvidia_smi(
        ("memory.total", "memory.used", "memory.free", "utilization.gpu"),
    )

    if not result:
        return {}

    return {
        "memory-used-GB/gpu": float(result["memory.used"]),
        "memory-free-GB/gpu": float(result["memory.free"]),
        "memory-percent/gpu": 100
        * float(result["memory.used"])
        / float(result["memory.total"]),
        "percent-usage/gpu": float(result["utilization.gpu"]),
    }


def _mps_log() -> dict[str, float]:
    """Return GPU information about current non-static status using ``sudo
    powermetrics``.

    Returns
    -------
        If ``sudo powermetrics`` is not executable (or is not configured for
        passwordless execution), returns ``None``, otherwise, we return a
        dictionary containing the following query information, in this order:

        * ``freq_hz`` as ``frequency-MHz/gpu``
        * 100 * (1 - ``idle_ratio``), as ``percent-usage/gpu``,
          (:py:class:`float`, in percent).
    """

    result = _run_powermetrics(500, key="gpu")

    if not result:
        return result

    return {
        "frequency-MHz/gpu": float(result["freq_hz"]),
        "percent-usage/gpu": 100 * (1 - result["idle_ratio"]),
    }


def cpu_constants() -> dict[str, int | float]:
    """Return static CPU information about the current system.

    Returns
    -------
    dict[str, int | float]
        A dictionary containing these entries:

        0. ``cpu_memory_total`` (:py:class:`float`): total memory available,
           in gigabytes
        1. ``cpu_count`` (:py:class:`int`): number of logical CPUs available.
    """

    return {
        "memory-total-GB/cpu": psutil.virtual_memory().total / GB,
        "number-of-cores/cpu": psutil.cpu_count(logical=True),
    }


class _CPUMonitor:
    """Monitors CPU information using :py:mod:`psutil`.

    Parameters
    ----------
    pid
        Process identifier of the main process (parent process) to observe.
    """

    def __init__(self, pid: int | None = None):
        this = psutil.Process(pid=pid)
        self.cluster = [this] + this.children(recursive=True)

        # touch cpu_percent() at least once for all processes in the cluster
        gone = set()
        for k in self.cluster:
            try:
                k.cpu_percent(interval=None)
            except (
                psutil.ZombieProcess,
                psutil.NoSuchProcess,
                psutil.AccessDenied,
            ):
                # child process is gone meanwhile
                # update the intermediate list for this time
                gone.add(k)
        self.cluster = list(set(self.cluster) - gone)

    def log(self) -> dict[str, int | float]:
        """Return current process cluster information.

        Returns
        -------
        dict[str, int | float]
            A dictionary containing these entries:

            0. ``cpu_memory_used`` (:py:class:`float`): total memory used from
               the system, in gigabytes
            1. ``cpu_rss`` (:py:class:`float`):  RAM currently used by
               process and children, in gigabytes
            2. ``cpu_vms`` (:py:class:`float`):  total memory (RAM + swap) currently
               used by process and children, in gigabytes
            3. ``cpu_percent`` (:py:class:`float`): percentage of the total CPU
               used by this process and children (recursively) since last call
               (first time called should be ignored).  This number depends on the
               number of CPUs in the system and can be greater than 100%
            4. ``cpu_processes`` (:py:class:`int`): total number of processes
               including self and children (recursively)
            5. ``cpu_open_files`` (:py:class:`int`): total number of open files by
               self and children
        """

        # check all cluster components and update process list
        # done so we can keep the cpu_percent() initialization
        stored_children = set(self.cluster[1:])
        current_children = set(self.cluster[0].children(recursive=True))
        keep_children = stored_children - current_children
        new_children = current_children - stored_children
        gone = set()
        for k in new_children:
            try:
                k.cpu_percent(interval=None)
            except (
                psutil.ZombieProcess,
                psutil.NoSuchProcess,
                psutil.AccessDenied,
            ):
                # child process is gone meanwhile
                # update the intermediate list for this time
                gone.add(k)
        new_children = new_children - gone
        self.cluster = self.cluster[:1] + list(keep_children) + list(new_children)

        memory_info = []
        cpu_percent = []
        open_files = []
        gone = set()

        for k in self.cluster:
            try:
                memory_info.append(k.memory_info())
                cpu_percent.append(k.cpu_percent(interval=None))
                open_files.append(len(k.open_files()))
            except (
                psutil.ZombieProcess,
                psutil.NoSuchProcess,
                psutil.AccessDenied,
            ):
                # child process is gone meanwhile, just ignore it
                # it is too late to update any intermediate list
                # at this point, but ensures to update counts later on
                gone.add(k)
        return {
            "memory-used-GB/cpu": psutil.virtual_memory().used / GB,
            "rss-GB/cpu": sum([k.rss for k in memory_info]) / GB,
            "vms-GB/cpu": sum([k.vms for k in memory_info]) / GB,
            "percent-usage/cpu": sum(cpu_percent),
            "num-processes/cpu": len(self.cluster) - len(gone),
            "num-open-files/cpu": sum(open_files),
        }

    def show_processes(self) -> str:
        """Return a string representation of currently running processes.

        Returns
        -------
            A string representation of currently running processes for
            debugging purposes.
        """
        return "\n".join([f"(pid={k.pid}) {k.cmdline()}" for k in self.cluster])


class _CombinedMonitor:
    """A helper monitor to log execution information.

    Parameters
    ----------
    device_type
        String representation of one of the supported pytorch device types
        triggering the correct readout of resource usage.
    main_pid
        The main process identifier to monitor.
    """

    def __init__(
        self,
        device_type: SupportedPytorchDevice,
        main_pid: int | None,
    ):
        self.cpu_monitor = _CPUMonitor(main_pid)
        self.device_type = device_type

    def log(self) -> dict[str, int | float]:
        """Combine measures from CPU and GPU monitors.

        Returns
        -------
            A dictionary containing the various logged information.
        """

        # we always log the CPU counters
        retval = {"timestamp": time.time()}
        retval.update(self.cpu_monitor.log())

        match self.device_type:
            case "cpu":
                pass
            case "cuda":
                entries = _cuda_log()
                if entries is not None:
                    retval.update(entries)
            case "mps":
                entries = _mps_log()
                if entries is not None:
                    retval.update(entries)
            case _:
                pass

        return retval


def _monitor_worker(
    interval: int | float,
    device_type: SupportedPytorchDevice,
    main_pid: int,
    stop_event: multiprocessing.synchronize.Event,
    wakeup_event: multiprocessing.synchronize.Event,
    queue: multiprocessing.Queue,
):
    """Monitor worker that measures resources and returns lists.

    Parameters
    ----------
    interval
        Number of seconds to wait between each measurement (maybe a floating
        point number as accepted by :py:func:`time.sleep`).
    device_type
        String representation of one of the supported pytorch device types
        triggering the correct readout of resource usage.
    main_pid
        The main process identifier to monitor.
    stop_event
        Event that indicates if we should continue running or stop.
    wakeup_event
        Event that indicates if we should log again immediately or wait for the
        interval timer.
    queue
        A queue, to send monitoring information back to the spawner.
    """

    monitor = _CombinedMonitor(device_type, main_pid)
    queue.put({})

    while not stop_event.is_set():
        try:
            start = time.time()
            queue.put(monitor.log())
            wait_at_most = interval - (time.time() - start)
            if wait_at_most < 0:
                warnings.warn(
                    f"The time to run a monitor.log() call "
                    f"({(time.time()-start):.2f}s) is larger than the "
                    f"interval of {interval}s. The amount of time between "
                    f"monitor log calls may be larger than the estimated "
                    f"interval time."
                )

            if wait_at_most > 0:
                wakeup_event.wait(timeout=wait_at_most)
            wakeup_event.clear()

        except Exception as e:
            import traceback

            warnings.warn(
                "Iterative CPU/GPU logging did not work properly. "
                "Traceback follows:\n"
                "".join(traceback.format_exception(type(e), e, e.__traceback__))
            )
            time.sleep(0.5)  # wait half a second, and try again!


class ResourceMonitor:
    """An external, non-blocking CPU/GPU resource monitor.

    Parameters
    ----------
    interval
        Number of seconds to wait between each measurement (maybe a floating
        point number as accepted by :py:func:`time.sleep`).
    device_type
        String representation of one of the supported pytorch device types
        triggering the correct readout of resource usage.
    main_pid
        The main process identifier to monitor.
    """

    def __init__(
        self,
        interval: int | float,
        device_type: SupportedPytorchDevice,
        main_pid: int,
    ):
        self.interval = interval
        self.device_type = device_type
        self.main_pid = main_pid
        self.stop_event = multiprocessing.Event()
        self.wakeup_event = multiprocessing.Event()
        self.q: multiprocessing.Queue[dict[str, int | float]] = multiprocessing.Queue()

        self.monitor = multiprocessing.Process(
            target=_monitor_worker,
            name="ResourceMonitorProcess",
            args=(
                self.interval,
                self.device_type,
                self.main_pid,
                self.stop_event,
                self.wakeup_event,
                self.q,
            ),
        )

    def start(self) -> None:
        """Start the monitoring process."""
        if not self.monitor.is_alive():
            timeout = 60  # seconds
            self.monitor.start()
            try:
                obj = self.q.get(timeout=timeout)
                if obj != {}:
                    raise RuntimeError(
                        f"Start of monitoring process did not work - we "
                        f"received a non-empty dictionary (size={len(obj)}) "
                        f"from the queue, instead of the expected `None`."
                    )
            except queue.Empty:
                raise RuntimeError(
                    f"Start of monitoring process did not work - we timed "
                    f"after {timeout}s waiting for a confirmation."
                )

    def __enter__(self) -> None:
        """Start the monitoring process."""
        self.start()

    def stop(self) -> None:
        """Stop the monitoring process."""

        if self.monitor.is_alive():
            self.stop_event.set()
            self.wakeup_event.set()
            self.monitor.join()
            if self.monitor.exitcode != 0:
                logger.error(
                    f"CPU/GPU resource monitor process exited with code "
                    f"{self.monitor.exitcode}.  Check logs for errors!",
                )

    def __exit__(self, *_) -> None:
        """Stop the monitoring process."""
        self.stop()

    def clear(self) -> None:
        """Prepare for an acquisition cycle by clearing logged events."""
        while not self.q.empty():
            self.q.get()
        self.wakeup_event.set()

    def checkpoint(self) -> list[dict[str, int | float]]:
        """Force the monitoring process to yield data and clear the internal
        accumulator.

        Returns
        -------
        list[dict[str, int | float]]
            A list of hardware counters that were monitored during the
            start/end interval.
        """

        data: list[dict[str, int | float]] = []
        while not self.q.empty():
            data.append(self.q.get())
        self.wakeup_event.set()
        return data


def aggregate(
    data: list[dict[str, int | float]],
    start: int | None = 0,
    end: int | None = None,
) -> dict[str, int | float]:
    """Aggregate monitored data to average/max each entry.

    Parameters
    ----------
    data
        Input data collected by the resource monitor, in the format of a list.
    start
        Start index to aggregate data from at the ``data`` list.
    end
        End index to aggregate data to at the ``data`` list.

    Returns
    -------
    dict[str, int | float]
        A collapsed representation of the input list with the common keys found
        in all dictionaries and averages (or maximum values) of measures found
        across similar keys.
    """
    if not data:
        return {}

    # aggregate data: list of dicts -> dict of lists
    def _acc(k, v):
        match k:
            case "num-processes/cpu" | "num-open-files/cpu":
                return numpy.max(v)
            case "timestamp":
                if len(v) > 1:
                    return v[-1] - v[0]
                return v[0]
            case _:
                return float(numpy.mean(v))

    retval = {
        k: _acc(k, [j.get(k, 0) for j in data][start:end]) for k in data[0].keys()
    }
    retval["cpu-monitoring-samples"] = len(data)
    return retval



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines most common types used in code."""

import abc
import collections.abc
import typing

import torch
import torch.utils.data

Sample: typing.TypeAlias = typing.Mapping[str, typing.Any]
"""Definition of a sample.

A dictionary containing an arbitrary number of keys and values.  Some of the
keys are reserved, others ignored within the framework, and can be re-used to
hold sample metadata required for further analysis.

Reserved keys:

* ``input``: This is typically a 1, 2 or 3D torch float tensor containing the
  input data to be analysed.
* ``target``: This is typically a torch float tensor containing the target the
  network must try to achieve. In the case of classification, it can be a 1D
  tensor containing a single entry (binary classification) or multiple entries
  (multi-class classification).  In the case of semantic segmentation, this
  entry typically contains a float representation of the target mask the
  network must decode from the ``input`` data.
* ``mask``: A torch float tensor containing a mask over which the input (and
  the output) may be ignored.  Typically used in semantic segmentation tasks.
* ``name``: A name for the sample.  Typically set to the name of the file or
  file-stem holding the ``input`` data.
"""


class RawDataLoader(abc.ABC):
    """A loader object can load samples from storage."""

    @abc.abstractmethod
    def sample(self, sample: typing.Any) -> Sample:
        """Load whole samples from media.

        Parameters
        ----------
        sample
            Information about the sample to load. Implementation dependent.

        Returns
        -------
            The instantiated sample, which is a dictionary where keys name the
            sample's data and metadata.
        """
        pass

    @abc.abstractmethod
    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """
        pass


Transform: typing.TypeAlias = typing.Callable[[torch.Tensor], torch.Tensor]
"""A callable that transforms tensors into (other) tensors.

Typically used in data-processing pipelines inside pytorch.
"""

TransformSequence: typing.TypeAlias = typing.Sequence[Transform]
"""A sequence of transforms."""

DatabaseSplit: typing.TypeAlias = collections.abc.Mapping[
    str,
    typing.Sequence[typing.Any],
]
"""The definition of a database split.

A database split maps dataset (subset) names to sequences of objects that,
through a :py:class:`RawDataLoader`, eventually becomes a :py:data:`.Sample` in
the processing pipeline.
"""

ConcatDatabaseSplit: typing.TypeAlias = collections.abc.Mapping[
    str,
    typing.Sequence[tuple[typing.Sequence[typing.Any], RawDataLoader]],
]
"""The definition of a complex database split composed of several other splits.

A database split maps dataset (subset) names to sequences of objects that,
through a :py:class:`.RawDataLoader`, eventually becomes a :py:data:`.Sample` in
the processing pipeline. Objects of this subtype allow the construction of
complex splits composed of cannibalized parts of other splits.  Each split may
be assigned a different :py:class:`.RawDataLoader`.
"""


class Dataset(torch.utils.data.Dataset[Sample], typing.Iterable, typing.Sized):
    """Our own definition of a pytorch Dataset.

    We iterate over Sample objects in this case.  Our datasets always
    provide a dunder len method.
    """

    def targets(self) -> list[torch.Tensor]:
        """Return the integer targets for all samples in the dataset."""
        raise NotImplementedError("You must implement the `targets()` method")


DataLoader: typing.TypeAlias = torch.utils.data.DataLoader[Sample]
"""Our own augmentation definition of a pytorch DataLoader.

We iterate over Sample objects in this case.
"""



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Simple affine augmentations for image analysis."""

import torchvision.transforms

augmentations = [
    torchvision.transforms.RandomAffine(
        degrees=10,
        translate=(0.1, 0.1),  # horizontal, vertical
        scale=(0.8, 1.0),  # minimum, maximum
        interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
    ),
]



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for TBX11K dataset."""

import importlib
import typing

import pytest
import torch
from click.testing import CliRunner


def id_function(val) -> str:
    match val:
        case str() | int():
            return str(val)
        case dict():
            return "{" + "".join(
                ",".join(
                    [id_function(k) + ":" + id_function(v) for k, v in val.items()]
                )
            )
        case tuple():
            return ",".join([id_function(k) for k in val])
        case _:
            return repr(val)


@pytest.mark.parametrize(
    "split,lengths,prefixes",
    [
        (
            "v1-healthy-vs-atb",
            dict(train=2767, validation=706, test=957),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-0",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-1",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-2",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-3",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-4",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-5",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-6",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-7",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-8",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v1-fold-9",
            dict(train=3177, validation=810, test=443),
            ("imgs/health", "imgs/tb"),
        ),
        (
            "v2-others-vs-atb",
            dict(train=5241, validation=1335, test=1793),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-0",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-1",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-2",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-3",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-4",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-5",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-6",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-7",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-8",
            dict(train=6003, validation=1529, test=837),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
        (
            "v2-fold-9",
            dict(train=6003, validation=1530, test=836),
            ("imgs/health", "imgs/sick", "imgs/tb"),
        ),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lengths: dict[str, int],
    prefixes: typing.Sequence[str],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.classify.data.tbx11k", f"{split}.json"),
        lengths=lengths,
        prefixes=prefixes,
        possible_labels=(0, 1),
    )


def check_loaded_batch(
    batch,
    batch_size: int,
    color_planes: int,
    prefixes: typing.Sequence[str],
    possible_labels: typing.Sequence[int],
    expected_num_labels: int,
    expected_image_shape: tuple[int, ...] | None = None,
):
    """Check the consistency of an individual (loaded) batch.

    Parameters
    ----------
    batch
        The loaded batch to be checked.
    batch_size
        The mini-batch size.
    color_planes
        The number of color planes in the images.
    prefixes
        Each file named in a split should start with at least one of these
        prefixes.
    possible_labels
        These are the list of possible labels contained in any split.
    expected_num_labels
        The expected number of labels each sample should have.
    expected_image_shape
        The expected shape of the image (num_channels, width, height).
    """

    # label, name and radiological sign bounding-boxes, image, bb targets
    assert len(batch) == 5

    assert isinstance(batch, dict)
    assert isinstance(batch["image"], torch.Tensor)
    assert batch["image"].shape[0] == batch_size  # mini-batch size
    assert batch["image"].shape[1] == color_planes
    assert batch["image"].shape[2] == batch["image"].shape[3]  # image is square

    if expected_image_shape:
        assert all([data.shape == expected_image_shape for data in batch["image"]])

    assert "target" in batch
    assert all([k in possible_labels for k in batch["target"]])

    if expected_num_labels:
        assert len(batch["target"]) == expected_num_labels

    assert "name" in batch
    assert all(
        [any([k.startswith(j) for j in prefixes]) for k in batch["name"]],
    )

    assert "bboxes" in batch

    for sample, target, bboxes, bbox_targets in zip(
        batch["image"],
        batch["target"],
        batch["bboxes"],
        batch["bbox_targets"],
    ):
        # there must be a sign indicated on the image, if active TB is detected
        if target[0] == 1:
            assert len(bboxes) != 0

        # eif label == 0:  # not true, may have TBI!
        #    assert len(bboxes) == 0

        # asserts all bounding boxes are within the raw image width and height
        if bboxes is not None:
            for bbox in bboxes:
                if target[0] == 1:
                    assert (bbox_targets == 1).all()
                else:
                    assert (bbox_targets == 0).all()
                assert (
                    bbox.data[2] < sample.shape[2]
                )  # check bbox is not outside image width
                assert bbox.data[3] < sample.shape[1]  # same with height
        else:
            assert bbox_targets is None

    # use the code below to view generated images
    # from torchvision.transforms.v2.functional import to_pil_image
    # to_pil_image(batch["image"]).show()
    # __import__("pdb").set_trace()


@pytest.mark.skip_if_rc_var_not_set("datadir.tbx11k")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["--limit=10", "tbx11k-v1-f0"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"

    result = runner.invoke(check, ["--limit=10", "tbx11k-v2-f0"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


@pytest.mark.skip_if_rc_var_not_set("datadir.tbx11k")
@pytest.mark.parametrize(
    "dataset",
    [
        "train",
        "validation",
        "test",
    ],
    ids=id_function,  # just changes how pytest prints it
)
@pytest.mark.parametrize(
    "name,prefixes",
    [
        ("v1_healthy_vs_atb", ("imgs/health", "imgs/tb")),
        ("v1_fold_0", ("imgs/health", "imgs/tb")),
        ("v1_fold_1", ("imgs/health", "imgs/tb")),
        ("v1_fold_2", ("imgs/health", "imgs/tb")),
        ("v1_fold_3", ("imgs/health", "imgs/tb")),
        ("v1_fold_4", ("imgs/health", "imgs/tb")),
        ("v1_fold_5", ("imgs/health", "imgs/tb")),
        ("v1_fold_6", ("imgs/health", "imgs/tb")),
        ("v1_fold_7", ("imgs/health", "imgs/tb")),
        ("v1_fold_8", ("imgs/health", "imgs/tb")),
        ("v1_fold_9", ("imgs/health", "imgs/tb")),
        ("v2_others_vs_atb", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_0", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_1", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_2", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_3", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_4", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_5", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_6", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_7", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_8", ("imgs/health", "imgs/sick", "imgs/tb")),
        ("v2_fold_9", ("imgs/health", "imgs/sick", "imgs/tb")),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_loading(name: str, dataset: str, prefixes: typing.Sequence[str]):
    datamodule = importlib.import_module(
        f".{name}", "mednet.config.classify.data.tbx11k"
    ).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.setup("predict")  # sets up all datasets

    loader = datamodule.predict_dataloader()[dataset]

    limit = 50  # limit load checking
    for batch in loader:
        if limit == 0:
            break
        check_loaded_batch(
            batch,
            batch_size=1,
            color_planes=3,
            prefixes=prefixes,
            possible_labels=(0, 1),
            expected_num_labels=1,
            expected_image_shape=(3, 512, 512),
        )
        limit -= 1


@pytest.mark.parametrize(
    "split",
    [
        "v1_fold_0",
        "v2_fold_0",
    ],
)
@pytest.mark.skip_if_rc_var_not_set("datadir.tbx11k")
def test_loaded_image_quality(database_checkers, datadir, split):
    reference_histogram_file = (
        datadir / f"histograms/raw_data/histograms_tbx11k_{split}.json"
    )

    datamodule = importlib.import_module(
        f".{split}", "mednet.config.classify.data.tbx11k"
    ).datamodule

    datamodule.model_transforms = []
    datamodule.batch_size = 2
    datamodule.setup("predict")

    database_checkers.check_image_quality(datamodule, reference_histogram_file)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Simple CNN for Tuberculosis Detection, to be trained from scratch.

Implementation of the model architecture proposed by F. Pasa in the article
"Efficient Deep Network Architectures for Fast Chest X-Ray Tuberculosis
Screening and Visualization".

Reference: :cite:p:`pasa_efficient_2019`
"""

import torch.nn
import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.classify.pasa
import mednet.models.transforms

model = mednet.models.classify.pasa.Pasa(
    loss_type=torch.nn.BCEWithLogitsLoss,
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=8e-5),
    model_transforms=[
        torchvision.transforms.v2.Grayscale(),
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
    ],
    augmentation_transforms=[],
)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""DRHAGIS dataset for Vessel Segmentation.

* Reference: :cite:p:`holm_dr_2017`

The DR HAGIS database has been created to aid the development of vessel
extraction algorithms suitable for retinal screening programmes. Researchers
are encouraged to test their segmentation algorithms using this database.

It should be noted that image 24 and 32 are identical, as this fundus image was
obtained from a patient exhibiting both diabetic retinopathy and age-related
macular degeneration.

The images resolutions (height x width) are one of:

* 4752x3168 px, or
* 3456x2304 px, or
* 3126x2136 px, or
* 2896x1944 px, or
* 2816x1880 px

* Protocol ``default``:

  * Training samples: 19 (including labels and masks)
  * Test samples: 20 (including labels and masks)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Drive dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = to_image(PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB"))
        image = to_dtype(image, torch.float32, scale=True)

        target = self.target(sample)

        mask = PIL.Image.open(self.datadir / sample[2]).convert(mode="1", dither=None)
        mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """DRHAGIS dataset for Vessel Segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import click
from clapper.click import verbosity_option

from .logging import setup_cli_logger

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup_cli_logger()


def _create_generic_figure(
    curves: dict[str, tuple[list[int], list[float]]],
    group: str,
) -> tuple:
    """Create a generic figure showing the evolution of a metric.

    This function will create a generic figure (one-size-fits-all kind of
    style) of a given metric across epochs.

    Parameters
    ----------
    curves
        A dictionary where keys represent all scalar names, and values
        correspond to a tuple that contains an array with epoch numbers (when
        values were taken), and the monitored values themselves.  These lists
        are pre-sorted by epoch number.
    group
        A scalar globs present in the existing tensorboard data that
        we are interested in for plotting.

    Returns
    -------
        A matplotlib figure and its axes.
    """

    import matplotlib.pyplot as plt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.ticker import MaxNLocator

    fig, ax = plt.subplots(1, 1)
    ax = typing.cast(Axes, ax)
    fig = typing.cast(Figure, fig)

    if len(curves) == 1:
        # there is only one curve, just plot it
        title, (epochs, values) = next(iter(curves.items()))
        ax.plot(epochs, values)

    else:
        # this is an aggregate plot, name things consistently
        labels = {k: k[len(group) - 1 :] for k in curves.keys()}
        title = group.rstrip("*").rstrip("/")
        for key, (epochs, values) in curves.items():
            ax.plot(epochs, values, label=labels[key])
        ax.legend(loc="best")

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_title(title)
    ax.set_xlabel("Epoch")
    ax.set_ylabel(title)

    ax.grid(alpha=0.3)
    fig.tight_layout()

    return fig, ax


def _create_loss_figure(
    curves: dict[str, tuple[list[int], list[float]]],
    group: str,
) -> tuple:
    """Create a specific figure of the loss evolution.

    This function will create a specific, more detailed figure of the loss
    evolution plot where the curves will be enriched with vertical lines
    indicating the points where the lowest validation losses are detected.  The
    higher the point oppacity, the lower is the loss on the validation set.

    Parameters
    ----------
    curves
        A dictionary where keys represent all scalar names, and values
        correspond to a tuple that contains an array with epoch numbers (when
        values were taken), and the monitored values themselves.  These lists
        are pre-sorted by epoch number.
    group
        A scalar globs present in the existing tensorboard data that
        we are interested in for plotting.

    Returns
    -------
        A matplotlib figure and its axes.
    """

    fig, ax = _create_generic_figure(curves, group)

    if "loss/validation" in curves:
        points = sorted(zip(*curves["loss/validation"]), key=lambda x: x[1])[:4]

        # create the highlights for each point, with fading colours
        alpha_decay = 0.5
        alpha = 1.0
        for x, y in points:
            ax.axvline(
                x=x,
                color="red",
                alpha=alpha,
                linestyle=":",
                label=f"epoch {x} (val={y:.3g})",
            )
            alpha *= 1 - alpha_decay

    ax.legend(loc="best")
    return fig, ax


def _create_figures(
    data: dict[str, tuple[list[int], list[float]]],
    groups: list[str] = [
        "loss/*",
        "learning-rate",
        "memory-used-GB/cpu/*rss-GB/cpu/*",
        "vms-GB/cpu/*",
        "num-open-files/cpu/*",
        "num-processes/cpu/*",
        "percent-usage/cpu/*",
        # nvidia gpu
        "memory-percent/gpu/*",
        "memory-used-GB/gpu/*",
        "memory-free-GB/gpu/*",
        "percent-usage/gpu/*",
    ],
) -> list:
    """Generate figures for each metric in the dataframe.

    Each row of the dataframe corresponds to an epoch and each column to a metric.
    It is assumed that some metric names are of the form <metric>/<subset>.
    All subsets for a metric will be displayed on the same figure.

    Parameters
    ----------
    data
        A dictionary where keys represent all scalar names, and values
        correspond to a tuple that contains an array with epoch numbers (when
        values were taken), and the monitored values themselves.  These lists
        are pre-sorted by epoch number.
    groups
        A list of scalar globs present in the existing tensorboard data that
        we are interested in for plotting.  Values with multiple matches are
        drawn on the same plot.  Values that do not exist are ignored.

    Returns
    -------
    list
        List of matplotlib figures, one per metric.
    """

    import fnmatch

    figures = []

    for group in groups:
        curves = {k: data[k] for k in fnmatch.filter(data.keys(), group)}

        if len(curves) == 0:
            continue

        if group == "loss/*":
            fig, _ = _create_loss_figure(curves, group)
            figures.append(fig)
        else:
            fig, _ = _create_generic_figure(curves, group)
            figures.append(fig)

    return figures


@click.command(
    epilog="""Examples:

\b
    1. Analyze a training log and produces various plots:

       .. code:: sh

          mednet train-analysis -vv -l results/logs
""",
)
@click.option(
    "--logdir",
    "-l",
    help="Path to the directory containing the Tensorboard training logs",
    required=True,
    type=click.Path(dir_okay=True, exists=True, path_type=pathlib.Path),
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    show_default=True,
    default="results",
)
@verbosity_option(logger=logger, expose_value=False)
def train_analysis(
    logdir: pathlib.Path,
    output_folder: pathlib.Path,
) -> None:  # numpydoc ignore=PR01
    """Create a plot for each metric in the training logs and saves them in a .pdf file."""
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages

    from ..utils.tensorboard import scalars_to_dict

    train_log_filename = "trainlog.pdf"
    train_log_file = pathlib.Path(output_folder) / train_log_filename

    data = scalars_to_dict(logdir)

    train_log_file.parent.mkdir(parents=True, exist_ok=True)

    with PdfPages(train_log_file) as pdf:
        for figure in _create_figures(data):
            pdf.savefig(figure)
            plt.close(figure)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`AlexNet network architecture <alexnet-pytorch_>`_, pre-trained with ImageNet weights.

.. note::

   The output layer is **always** initialized from scratch.
"""

import torch.nn
import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.classify.alexnet
import mednet.models.transforms

model = mednet.models.classify.alexnet.Alexnet(
    loss_type=torch.nn.BCEWithLogitsLoss,
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=0.0001),
    pretrained=True,
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        # Alexnet requires a specific input size that works for its avgpool
        # layer, which expects to downsample the input size to a multiple of 6.
        # Empirical tests show that if the input size is ..., then the
        # avgpooling layer will get these number of features:
        # 414 x 414 px -> 11 features (not OK)
        # 415 x 415 px -> 12 features (OK)
        # 446 x 446 px -> 12 features (OK)
        # 447 x 447 px -> 13 features (not OK)
        torchvision.transforms.v2.Resize(430, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Test code for losses."""

import numpy
import pytest
import torch

from mednet.models.losses import MOONBCEWithLogitsLoss


@pytest.mark.skip_if_rc_var_not_set("datadir.montgomery")
def test_pos_weight_for_bcewithlogitsloss_binary():
    from mednet.config.classify.data.montgomery.default import datamodule
    from mednet.models.losses import pos_weight_for_bcewithlogitsloss

    datamodule.model_transforms = []
    datamodule.prepare_data()
    datamodule.setup(stage="fit")

    train, val = pos_weight_for_bcewithlogitsloss(datamodule)

    assert "pos_weight" in train
    assert train["pos_weight"].shape == (1,)
    # there are 37 positives out of 88 samples in the training set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (88 - 37) / 37
    assert numpy.isclose(train["pos_weight"].item(), pos_ratio)

    assert "pos_weight" in val
    assert train["pos_weight"].shape == (1,)
    # there are 9 positives out of 22 samples in the validation set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (22 - 9) / 9
    assert numpy.isclose(val["pos_weight"].item(), pos_ratio)


@pytest.mark.skip_if_rc_var_not_set("datadir.montgomery")
def test_pos_weight_for_bcewithlogitsloss_multiclass():
    from mednet.config.classify.data.montgomery.multiclass import datamodule
    from mednet.models.losses import pos_weight_for_bcewithlogitsloss

    datamodule.model_transforms = []
    datamodule.prepare_data()
    datamodule.setup(stage="fit")

    train, val = pos_weight_for_bcewithlogitsloss(datamodule)

    assert "pos_weight" in train
    assert train["pos_weight"].shape == (2,)
    # there are 37 positives out of 88 samples in the training set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (88 - 37) / 37
    assert numpy.allclose(train["pos_weight"], (1 / pos_ratio, pos_ratio))

    assert "pos_weight" in val
    assert train["pos_weight"].shape == (2,)
    # there are 9 positives out of 22 samples in the validation set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (22 - 9) / 9
    assert numpy.allclose(val["pos_weight"], (1 / pos_ratio, pos_ratio))


@pytest.mark.skip_if_rc_var_not_set("datadir.montgomery")
def test_weight_for_moonbcewithlogitsloss_weights():
    from mednet.config.classify.data.montgomery.multiclass import datamodule

    datamodule.model_transforms = []
    datamodule.prepare_data()
    datamodule.setup(stage="fit")

    train, val = MOONBCEWithLogitsLoss.get_arguments_from_datamodule(datamodule)

    assert "weights" in train
    assert train["weights"].shape == (2, 2)
    # there are 37 positives out of 88 samples in the training set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (88 - 37) / 37
    # the w- vector should be [1 , 1 / pos_ratio] (1st row)
    # the w+ vector should be [1/pos_ratio , 1] (2nd row)
    assert numpy.allclose(train["weights"], ((1, 1 / pos_ratio), (1 / pos_ratio, 1)))

    assert "weights" in val
    assert train["weights"].shape == (2, 2)
    # there are 9 positives out of 22 samples in the validation set
    # therefore, pos_weight = # neg / # pos
    pos_ratio = (22 - 9) / 9
    # the w- vector should be [1 , 1/pos_ratio] (1st row)
    # the w+ vector should be [1/pos_ratio , 1] (2nd row)
    assert numpy.allclose(val["weights"], ((1, 1 / pos_ratio), (1 / pos_ratio, 1)))


def test_weight_for_moonbcewithlogitsloss_multiclass():
    target = torch.FloatTensor(
        [
            [0, 1],
            [0, 1],
            [1, 0],
            [1, 0],
            [1, 0],
        ]
    )
    model_output = torch.FloatTensor(
        [
            [0, 1],  # ideal case
            [1, 0],  # worst case
            [0.5, 0.5],
            [0.75, 0.10],
            [0.9, 0.25],
        ]
    )

    # We should first create a composition of weights depending on the "ones"
    # in each column. If there is a "1" on a target item, then it should get
    # the w+ (position [1]) of the weight of that column, otherwise, the w-
    # (position [0]). Manually, this should be equivalent to this:
    pos_ratio = (88 - 37) / 37  # copied from previous test
    w = torch.FloatTensor(((1, 1 / pos_ratio), (1 / pos_ratio, 1)))
    w_neg, w_pos = w
    weights_manual = torch.FloatTensor(
        [
            # N.B.: weights are organized as a [2, C] matrix with w- at row [0]
            # and w+ at row [1]
            [w_neg[0], w_pos[1]],
            [w_neg[0], w_pos[1]],
            [w_pos[0], w_neg[1]],
            [w_pos[0], w_neg[1]],
            [w_pos[0], w_neg[1]],
        ]
    )

    # You can implement this in a faster operation using the following
    # expression:
    weights = ((1 - target) * w_neg) + (target * w_pos)
    assert numpy.allclose(weights_manual, weights)

    # the loss should evaluate the BCE with logits first
    individual_losses = torch.nn.functional.binary_cross_entropy_with_logits(
        model_output, target, reduction="none"
    )

    loss = MOONBCEWithLogitsLoss(weights=w)
    result = loss.forward(model_output, target)
    assert numpy.isclose(result, (individual_losses * weights).mean())


def test_compare_moon_to_bce_loss():
    # Tests that MOON and BCE with simple weighting are different only w.r.t.
    # the weights that are used to weigh samples.

    pos_bce_weight = torch.FloatTensor([3, 0.5, 1])

    loss_bce = torch.nn.BCEWithLogitsLoss(pos_weight=pos_bce_weight)

    # MOON: neg_moon_weight = (1 / pos_bce_weight).clamp(min=0, max=1)
    neg_moon_weight = torch.ones_like(pos_bce_weight)  # BCE equivalent
    # MOON: pos_moon_weight = pos_bce_weight.clamp(min=0, max=1)
    pos_moon_weight = pos_bce_weight  # BCE equivalent
    moon_weights = torch.vstack([neg_moon_weight, pos_moon_weight])
    loss_moon = MOONBCEWithLogitsLoss(weights=moon_weights)

    # assert they provide the same answer
    target = torch.FloatTensor(
        [
            [0, 1, 1],
            [0, 1, 0],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 1],
        ]
    )
    model_output = torch.FloatTensor(
        [
            [0, 1, 1],  # ideal case
            [1, 0, 1],  # worst case
            [0.5, 0.5, 0.5],
            [0.75, 0.10, 0.1],
            [0.9, 0.25, 0.8],
        ]
    )

    assert loss_bce.forward(model_output, target) == loss_moon.forward(
        model_output, target
    )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Japanese Society of Radiological Technology dataset for lung segmentation.

The database includes 154 nodule and 93 non-nodule images.  It contains a total
of 247 resolution of 2048 x 2048 pixels, issued from original digitized
Radiographies (laser scanner). One set of ground-truth lung annotations is
available.

* Database references:

  * Original data: :cite:p:`shiraishi_development_2000`
  * Split: :cite:p:`gaal_attention_2020`

.. important:: **Raw data organization**

   The JSRT_ base datadir, which you should configure following the
   :ref:`mednet.setup` instructions, must contain at least the following
   directories:

   - ``All247images/`` (directory containing the CXR images, in raw format)
   - ``scratch/`` (must contain masks downloaded from `JSRT-Annotations`_)

Data specifications:

* Raw data input (on disk):

  * Original images encoded in proprietary 12-bit RAW format.  A PNG-converted
    set of images is provided at JSRT-Kaggle_ for your reference.  Input
    resolution is 2048 x 2048 pixels.
  * Masks: encoded as GIF files with separate portions for left and right
    lungs, with a resolution of 1024 x 1024 pixels
  * Total samples: 247

* Output sample:

    * Image: Load raw image from folder ``All247images/`` using
      :py:func:`numpy.fromfile`, then applies a simple histogram equalization
      to the 8-bit representation of the image, to obtain something along the
      lines of the PNG (unofficial) version distributed at JSRT-Kaggle_.
      Output images have a size of 1024 x 1024 pixels, achieved by resizing the
      original input with bilinear interpolation.
    * Labels for each of the lungs are read from the provided GIF files and
      merged into a single output image.

The ``default`` split contains 172 samples for training, 25 for validation and
50 for test.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import numpy as np
import PIL.Image
import skimage.exposure
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the jsrt dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def load_pil_raw_12bit_jsrt(self, path: pathlib.Path) -> PIL.Image.Image:
        """Load a raw 16-bit sample data.

        This method was designed to handle the raw images from the JSRT dataset.
        It reads the data file and applies a simple histogram equalization to the
        8-bit representation of the image to obtain something along the lines of
        the PNG (unofficial) version distributed at `JSRT-Kaggle`.

        Parameters
        ----------
        path
            The full path leading to the image to be loaded.

        Returns
        -------
            A PIL image in RGB mode, with `width`x`width` pixels.
        """

        raw_image = np.fromfile(path, np.dtype(">u2")).reshape(2048, 2048)
        raw_image[raw_image > 4095] = 4095
        raw_image = 4095 - raw_image  # invert colors
        raw_image = (raw_image >> 4).astype(np.uint8)  # 8-bit uint
        raw_image = skimage.exposure.equalize_hist(raw_image)
        return PIL.Image.fromarray((raw_image * 255).astype(np.uint8)).convert("RGB")

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = self.load_pil_raw_12bit_jsrt(self.datadir / sample[0])

        # N.B.: Original image size is 2048x2048 px, however annotations were done on
        # 1024x1024 px canvas.
        #
        # assert image.size == (2048, 2048)

        image = image.resize((1024, 1024), PIL.Image.Resampling.BILINEAR)
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        # use image as a base since target() can be overriden by child class
        mask = torch.ones((1, image.shape[-2], image.shape[-1]), dtype=torch.float32)

        image = tv_tensors.Image(image)
        target = tv_tensors.Mask(target)
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        # Combine left and right lung masks into a single tensor
        left = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        right = PIL.Image.open(self.datadir / sample[2]).convert(mode="1", dither=None)
        target = np.ma.mask_or(np.asarray(left), np.asarray(right))

        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """Japanese Society of Radiological Technology dataset for lung segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# Adapted from https://github.com/pytorch/pytorch/issues/2001#issuecomment-405675488

from functools import reduce

import torch
from torch.nn.modules.module import _addindent


# ignore this space!
def _repr(model: torch.nn.Module) -> tuple[str, int]:
    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    extra_repr = model.extra_repr()
    # empty string will be split into list ['']
    if extra_repr:
        extra_lines = extra_repr.split("\n")
    child_lines = []
    total_params = 0
    for key, module in model._modules.items():  # noqa: SLF001
        mod_str, num_params = _repr(module)
        mod_str = _addindent(mod_str, 2)
        child_lines.append("(" + key + "): " + mod_str)
        total_params += num_params
    lines = extra_lines + child_lines

    for _, p in model._parameters.items():  # noqa: SLF001
        if hasattr(p, "dtype"):
            total_params += reduce(lambda x, y: x * y, p.shape)

    main_str = model._get_name() + "("  # noqa: SLF001
    if lines:
        # simple one-liner info, which most builtin Modules will use
        if len(extra_lines) == 1 and not child_lines:
            main_str += extra_lines[0]
        else:
            main_str += "\n  " + "\n  ".join(lines) + "\n"

    main_str += ")"
    main_str += f", {total_params:,} params"
    return main_str, total_params


def summary(model: torch.nn.Module) -> tuple[str, int]:
    """Count the number of parameters in each model layer.

    Parameters
    ----------
    model
        Model to summarize.

    Returns
    -------
    tuple[int, str]
        A tuple containing a multiline string representation of the network and the number of parameters.
    """
    return _repr(model)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Simple CNN for 3D image classification, to be trained from scratch."""

from torch.nn import BCEWithLogitsLoss
from torch.optim import Adam

from mednet.models.classify.cnn3d import Conv3DNet

model = Conv3DNet(
    loss_type=BCEWithLogitsLoss,
    optimizer_type=Adam,
    optimizer_arguments=dict(lr=8e-5),
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Mobile2 UNet network architecture, from :cite:p:`laibacher_m2u-net_2018`."""

import logging
import typing

import torch
import torch.nn
import torch.utils.data
from torchvision.models.mobilenetv2 import InvertedResidual

from ...data.typing import TransformSequence
from ..losses import SoftJaccardAndBCEWithLogitsLoss
from .backbones.mobilenetv2 import mobilenet_v2_for_segmentation
from .model import Model

logger = logging.getLogger(__name__)


class DecoderBlock(torch.nn.Module):
    """Decoder block: upsample and concatenate with features maps from the
    encoder part.

    Parameters
    ----------
    up_in_c
        Number of input channels.
    x_in_c
        Number of cat channels.
    upsamplemode
        Mode to use for upsampling.
    expand_ratio
        The expand ratio.
    """

    def __init__(self, up_in_c, x_in_c, upsamplemode="bilinear", expand_ratio=0.15):
        super().__init__()
        self.upsample = torch.nn.Upsample(
            scale_factor=2, mode=upsamplemode, align_corners=False
        )  # H, W -> 2H, 2W
        self.ir1 = InvertedResidual(
            up_in_c + x_in_c,
            (x_in_c + up_in_c) // 2,
            stride=1,
            expand_ratio=expand_ratio,
        )

    def forward(self, up_in, x_in):
        up_out = self.upsample(up_in)
        cat_x = torch.cat([up_out, x_in], dim=1)
        return self.ir1(cat_x)


class LastDecoderBlock(torch.nn.Module):
    """Last decoder block.

    Parameters
    ----------
    x_in_c
        Number of cat channels.
    upsamplemode
        Mode to use for upsampling.
    expand_ratio
        The expand ratio.
    """

    def __init__(self, x_in_c, upsamplemode="bilinear", expand_ratio=0.15):
        super().__init__()
        self.upsample = torch.nn.Upsample(
            scale_factor=2, mode=upsamplemode, align_corners=False
        )  # H, W -> 2H, 2W
        self.ir1 = InvertedResidual(x_in_c, 1, stride=1, expand_ratio=expand_ratio)

    def forward(self, up_in, x_in):
        up_out = self.upsample(up_in)
        cat_x = torch.cat([up_out, x_in], dim=1)
        return self.ir1(cat_x)


class M2UNetHead(torch.nn.Module):
    """M2U-Net head module.

    Parameters
    ----------
    in_channels_list
        Number of channels for each feature map that is returned from backbone.
    upsamplemode
        Mode to use for upsampling.
    expand_ratio
        The expand ratio.
    """

    def __init__(
        self, in_channels_list=None, upsamplemode="bilinear", expand_ratio=0.15
    ):
        super().__init__()

        # Decoder
        self.decode4 = DecoderBlock(96, 32, upsamplemode, expand_ratio)
        self.decode3 = DecoderBlock(64, 24, upsamplemode, expand_ratio)
        self.decode2 = DecoderBlock(44, 16, upsamplemode, expand_ratio)
        self.decode1 = LastDecoderBlock(33, upsamplemode, expand_ratio)

        # initilaize weights
        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.kaiming_uniform_(m.weight, a=1)
                if m.bias is not None:
                    torch.nn.init.constant_(m.bias, 0)
            elif isinstance(m, torch.nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        decode4 = self.decode4(x[5], x[4])  # 96, 32
        decode3 = self.decode3(decode4, x[3])  # 64, 24
        decode2 = self.decode2(decode3, x[2])  # 44, 16
        return self.decode1(decode2, x[1])  # 30, 3


class M2Unet(Model):
    """Mobile2 UNet network architecture, from :cite:p:`laibacher_m2u-net_2018`.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    pretrained
        If True, will use VGG16 pretrained weights.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = SoftJaccardAndBCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
    ):
        super().__init__(
            name="m2unet",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=1,  # fixed at current implementation
        )

        if pretrained:
            from ..normalizer import make_imagenet_normalizer

            Model.normalizer.fset(self, make_imagenet_normalizer())  # type: ignore[attr-defined]

        self.backbone = mobilenet_v2_for_segmentation(
            pretrained=pretrained,
            return_features=[1, 3, 6, 13],
        )

        self.head = M2UNetHead(in_channels_list=[16, 24, 32, 96])

    def forward(self, x):
        x = self.normalizer(x)
        x = self.backbone(x)
        return self.head(x)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from clapper.rc import UserDefaults


def load_rc() -> UserDefaults:
    """Return global configuration variables.

    Returns
    -------
        The user defaults read from the user .toml configuration file.
    """
    return UserDefaults("mednet.toml")
